# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from string import hexdigits as HEXDIGITS
from collections import abc
import os
import base64
import unicodedata

from ..utils import Response
from .torrent import TorrentFields, Torrent
from .. import ClientError
from ..filters.tfilter import TorrentFilter
from ..filters.ffilter import TorrentFileFilter


class _TorrentCache():
    def __init__(self, raw_torrents=()):
        self._tdict = {}  # Map torrent IDs to Torrent objects

    def update(self, raw_torrents):
        import time ; start = time.time()

        rtdict = {rt['id']:rt for rt in raw_torrents}
        tdict = self._tdict

        # Add or update torrents
        for tid,t in rtdict.items():
            if tid in tdict:
                # log.debug('Updating torrent #%d', tid)
                tdict[tid].update(t)
            else:
                # log.debug('Adding torrent #%d', tid)
                tdict[tid] = Torrent(t)

        log.debug('Updated %d cached torrents in %.3fms',
                  len(tdict), (time.time()-start)*1000)

    def purge(self, existing_tids):
        """Remove torrents with IDs that are not in `existing_ids`"""
        tdict = self._tdict
        known_tids = set(tdict)
        removed_tids = known_tids.difference(existing_tids)
        for tid in removed_tids:
            del tdict[tid]

    def get(self, *ids):
        """Return tuple of Torrent objects"""
        if ids:
            return tuple(t for tid,t in self._tdict.items() if tid in ids)
        else:
            return tuple(t for t in self._tdict.values())

    def files_initialized(self, ids):
        """Wether cached Torrents have a 'files' key"""
        return all('files' in t
                   for t in self._tdict.values()
                   if t['id'] in ids)

    def __len__(self):
        return len(self._tdict)

    def __repr__(self):
        tlist = ', '.join('#'+str(tid) for tid in self._tdict)
        return '<{cls} {tlist}>'.format(cls=type(self).__name__,
                                        tlist=tlist or '(empty)')


class TorrentAPI():
    """High-level abstraction of the Transmission RPC protocol"""

    def __init__(self, rpc):
        self.rpc = rpc
        self._tcache = _TorrentCache()

    async def _request(self, method, *args, **kwargs):
        try:
            result = await method(*args, **kwargs)
        except ClientError as e:
            return Response(result=None, msgs=(e,), success=False)
        else:
            return Response(result=result, msgs=(), success=True)

    async def add(self, torrent, stopped=False):
        """Add torrent from file, URL or hash

        torrent: Path to local file, web/magnet link or hash
        stopped: False to start downloading immediately, True otherise

        Return Response with the following properties:
            torrent: Torrent object with the keys 'id' and 'name' if the
                     torrent could be added or if it already exists, otherwise
                     None
            success: False if torrent could not be added or already exists,
                     True otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        torrent_str = torrent
        torrent = None
        msgs = []
        success = False

        args = {'paused': stopped}

        # Check if torrent_str is path to local torrent file
        torrent_str_path = os.path.expanduser(torrent_str)
        if os.path.exists(torrent_str_path):
            torrent_str = torrent_str_path
            del torrent_str_path

            # Read local file
            try:
                with open(torrent_str, 'rb') as f:
                    args['metainfo'] = str(base64.b64encode(f.read()),
                                           encoding='ascii')
            except OSError as e:
                msgs.append(ClientError('%s: %s' % (e.strerror, torrent_str)))
                return Response(success=success, torrent=torrent, msgs=msgs)
        elif len(torrent_str) == 40 and all(c in HEXDIGITS for c in torrent_str):
            # Convert hash to magnet link
            args['filename'] = 'magnet:?xt=urn:btih:' + torrent_str
        else:
            # It's probably a link, let the daemon figure it out
            args['filename'] = torrent_str

        response = await self._request(self.rpc.torrent_add, **args)
        if not response.success:
            if 'Invalid or corrupt' in str(response.msgs[0]):
                msgs = (ClientError('Invalid or corrupt torrent: {!r}'.format(torrent_str)),)
            else:
                msgs = response.msgs
            return Response(success=False, torrent=None, msgs=msgs)
        else:
            result = response.result
            if 'torrent-duplicate' in result:
                info = result['torrent-duplicate']
                msgs.append(ClientError('Torrent already exists: ' + info['name']))
                success = False
            elif 'torrent-added' in result:
                info = result['torrent-added']
                msgs.append('Added %s' % info['name'])
                success = True
            else:
                raise RuntimeError('Malformed response: {}'.format(result))
            torrent = Torrent({'id': info['id'], 'name': info['name']})

        return Response(success=success, torrent=torrent, msgs=msgs)


    async def _request_torrents(self, fields, ids=None):
        """Unmodified 'torrent-get' request"""
        try:
            if 'id' not in fields:
                fields = ('id',) + tuple(fields)
            if ids is None:
                # Request all IDs
                raw_tlist = await self.rpc.torrent_get(fields=fields)
            else:
                if len(ids) > 0:
                    # Request given IDs
                    raw_tlist = await self.rpc.torrent_get(fields=fields, ids=ids)
                else:
                    # No IDs (i.e. empty torrent list) requested
                    raw_tlist = []
        except ClientError as e:
            return Response(success=False, raw_torrents=[], msgs=[e])
        else:
            self._tcache.update(raw_tlist)
            return Response(success=True, raw_torrents=raw_tlist)

    async def _get_torrents_by_ids(self, keys, ids=None, autoconnect=True):
        """Return a Response object with 'torrents' set to a tuple of Torrents

        keys: 'ALL' for all supported Torrent keys or a sequence of key
              strings (see client.tkeys.TYPES for available keys)
        ids: None for all torrents or a sequence of wanted IDs
        autoconnect: Wether to attempt to connect automatically if not
                     connected; if False and not connected, return None
        """
        if not autoconnect and not self.rpc.connected:
            return None

        # The 'files' RPC field only returns static information (variable
        # information is in 'filestats'), and since all Torrent values are
        # cached anyway, we only have to request it once.
        if 'files' in keys and not self._tcache.files_initialized(ids):
            log.debug('Initializing files for torrents: %s', ids)
            response = await self._request_torrents(('files',), ids)
            if not response.success:
                return Response(success=False, torrents=(), msgs=response.msgs)

        if keys == 'ALL':
            fields = TorrentFields(keys)
        else:
            fields = TorrentFields(*keys)
        tlist = ()
        msgs = []
        success = False

        response = await self._request_torrents(fields, ids)
        if not response.success:
            return Response(success=False, torrents=(), msgs=response.msgs)
        else:
            from time import time
            start = time()

            raw_tlist = response.raw_torrents
            if ids:
                tlist = self._tcache.get(*ids)
                for tid in ids:
                    # Torrent objects are equal to an integer of the torrent's ID
                    if tid not in tlist:
                        msgs.append(ClientError('No torrent with ID: {}'.format(tid)))
            else:
                self._tcache.purge(t['id'] for t in raw_tlist)  # Remove deleted torrents from cache
                tlist = self._tcache.get()
            success = len(tlist) > 0 or not ids

            log.debug('Found %d torrents in %.3fms', len(tlist), (time()-start)*1e3)

        return Response(success=success, torrents=tlist, msgs=msgs)

    async def _get_torrents_by_filter(self, keys, tfilter=None, autoconnect=True):
        """Return a Response object with 'torrents' set to a tuple of Torrents

        keys: See _get_torrents_by_ids
        tfilter: A TorrentFilter instance or None
        autoconnect: See _get_torrents_by_ids
        """
        if not autoconnect and not self.rpc.connected:
            return None
        elif tfilter == None:
            log.debug('Looking for all torrents with keys: %s', keys)
            # No filter specified - just return all torrents with the specified keys
            return await self._get_torrents_by_ids(keys=keys, autoconnect=autoconnect)
        else:
            log.debug('Looking for %s torrents with keys: %s', tfilter, keys)
            tlist = ()
            msgs = []
            success = False
            if isinstance(tfilter, str):
                tfilter = TorrentFilter(tfilter)

            # Request all torrents with the keys needed to filter them
            log.debug('Requesting full list with filter keys: %s', tfilter.needed_keys)
            response = await self._get_torrents_by_ids(keys=tfilter.needed_keys)
            if response.success:
                # Find IDs of torrents that match tfilter
                wanted_ids = tuple(t['id'] for t in tfilter.apply(response.torrents))
                log.debug('Wanted IDs: %s', wanted_ids)
                if len(wanted_ids) > 0:
                    # Get only wanted torrents with all wanted keys
                    try:
                        response = await self._get_torrents_by_ids(keys, wanted_ids)
                    except ClientError as e:
                        msgs.append(ClientError(str(e)))
                    else:
                        tlist = tuple(response.torrents)
            else:
                msgs.extend(response.msgs)

            success = len(tlist) > 0
            if not success:
                msgs.append(ClientError('No matching torrents: {}'.format(tfilter)))
            else:
                msgs.append('Found {} {} torrent{}'.format(
                    len(tlist), tfilter, '' if len(tlist) == 1 else 's'))

            return Response(success=success, torrents=tlist, msgs=msgs)

    async def torrents(self, torrents=None, keys='ALL', autoconnect=True):
        """Fetch and return torrents

        torrents: Iterator of torrent IDs, TorrentFilter object (or its string
                  representation) or None for all torrents
        keys: tuple of Torrent keys to fetch or 'ALL' for all torrents
        autoconnect: Wether to attempt to connect automatically if not
                     connected; if False and not connected, return None

        Return Response with the following properties:
            torrents: tuple of Torrent objects with requested torrents
            success: False if no torrents were found, True otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        if torrents is None:
            return await self._get_torrents_by_ids(keys, autoconnect=autoconnect)
        elif isinstance(torrents, (str, TorrentFilter)):
            return await self._get_torrents_by_filter(keys, tfilter=torrents, autoconnect=autoconnect)
        elif isinstance(torrents, abc.Sequence) and \
             all(isinstance(id, int) for id in torrents):
            return await self._get_torrents_by_ids(keys, ids=torrents, autoconnect=autoconnect)
        else:
            raise ValueError("Invalid 'torrents' argument: {!r}".format(torrents))

    async def _torrent_action(self, method, torrents=None, method_args={}, check=None,
                              keys=(), autoconnect=True):
        """Helper method that operates on torrents (start, stop, remove, etc)

        method: Any method from TransmissionRPC that accepts torrent ids
        torrents: See `torrents` method
        method_args: Dictionary with keyword arguments for method (except 'ids')
        check: None or callable that is called with every torrent; must return
               a 2-tuple of (SUCCESS, MESSAGE) where SUCCESS is evaluated as
               bool and MESSAGE a string or None.  If SUCCESS evaluates to
               True, `method` is applied to the torrent, otherwise not.
        keys: List of Torrent keys the check function needs
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents that `method` was applied to with the
                      keys 'id' and 'name'
            success: True if `method` was successfully applied to at least one
                     torrent, False otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        if not autoconnect and not self.rpc.connected:
            return None

        tlist = []
        msgs = []
        success = False

        response = await self.torrents(torrents, keys=keys)
        if not response.success:
            return Response(success=False, torrents=(), msgs=response.msgs)
        else:
            msgs = list(response.msgs)

        if check is None:
            tlist = response.torrents
        else:
            # Filter torrents through check function
            for t in response.torrents:
                passed, msg = check(t)
                if passed:
                    tlist.append(t)
                    if msg is not None:
                        msgs.append(msg)
                else:
                    if msg is not None:
                        msgs.append(ClientError(msg))

        # Apply method to torrents that passed the check function
        if len(tlist) > 0:
            try:
                # Ignore response because it is always {}, except for
                # 'torrent-get' requests, which this method is not meant for.
                await method(ids=tuple(t['id'] for t in tlist), **method_args)
            except ClientError as e:
                msgs.append(e)
                tlist = ()

        # The resulting torrents have only 'id' and 'name' keys
        tlist = tuple(Torrent({'id': t['id'], 'name': t['name']})
                      for t in tlist)
        success = len(tlist) > 0
        return Response(success=success, torrents=tlist, msgs=msgs)

    async def stop(self, torrents, autoconnect=True):
        """Stop down-/uploading torrents

        torrents: See `torrents` method
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents that were stopped with the keys 'id'
                      and 'name'
            success: True if any torrents were found and stopped, False otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        def check(t):
            if t['status'] == 'stopped':
                return (False, 'Already stopped: ' + t['name'])
            else:
                return (True, 'Stopping ' + t['name'])

        return await self._torrent_action(self.rpc.torrent_stop, torrents,
                                          check=check, keys=('status', 'name'),
                                          autoconnect=autoconnect)

    async def start(self, torrents, force=False, autoconnect=True):
        """Start down-/uploading torrents

        torrents: See `torrents` method
        force: Start downloading even if download queue is active and full
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents that were started with the keys 'id'
                      and 'name'
            success: True if any torrents were found and started, False
                     otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        def check(t):
            if t['status'] == 'stopped':
                return (True, 'Starting ' + t['name'])
            else:
                return (False, 'Already started: ' + t['name'])

        if force:
            method = self.rpc.torrent_start_now
        else:
            method = self.rpc.torrent_start

        return await self._torrent_action(method, torrents,
                                          method_args={'force':force},
                                          check=check, keys=('status', 'name'),
                                          autoconnect=autoconnect)

    async def toggle_stopped(self, torrents, force=False, autoconnect=True):
        """Start down-/uploading torrents

        torrents: See `torrents` method
        force: See `start` method
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents that were toggled with the keys 'id'
                      and 'name'
            success: True if any torrents were found, False otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        response = await self.torrents(torrents, keys=('status',))
        if not response.success:
            return Response(success=False, torrents=(), msgs=response.msgs)

        stopped, running = [], []
        for t in response.torrents:
            if t['status'] == 'stopped':
                stopped.append(t)
            else:
                running.append(t)

        torrents, msgs = ((), [])
        if len(running) > 0:
            r = await self.stop(tuple(t['id'] for t in running))
            torrents += r.torrents
            msgs += r.msgs
        if len(stopped) > 0:
            r = await self.start(tuple(t['id'] for t in stopped), force=force)
            torrents += r.torrents
            msgs += r.msgs

        return Response(torrents=torrents,
                        success=len(torrents) > 0,
                        msgs=msgs)

    async def verify(self, torrents, autoconnect=True):
        """Verify torrents's downloaded data

        torrents: See `torrents` method
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents that will be verified with the keys
                      'id' and 'name'
            success: True if any torrents were found and will be verified,
                     False otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        def check(t):
            if t['status'] == 'verifying':
                return (False, 'Already verifying: ' + t['name'])
            elif t['status'] == 'verifying pending':
                return (False, 'Already queued for verification: ' + t['name'])
            else:
                return (True, 'Verifying ' + t['name'])

        return await self._torrent_action(self.rpc.torrent_verify, torrents,
                                          check=check, keys=('status', 'name'),
                                          autoconnect=autoconnect)

    async def remove(self, torrents, delete=False, autoconnect=True):
        """Remove torrents

        torrents: See `torrents` method
        delete: True if downloaded files should be deleted
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents that were removed with the keys 'id'
                      and 'name'
            success: True if any torrents were found and removed, False
                     otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        if delete:
            msg = 'Deleting %s (including files)'
        else:
            msg = 'Removing %s (keeping files)'

        def create_info_msg(t):
            return (True, msg % t['name'])

        return await self._torrent_action(self.rpc.torrent_remove, torrents,
                                          keys=('name',),
                                          method_args={'delete-local-data': delete},
                                          check=create_info_msg,
                                          autoconnect=autoconnect)


    async def move(self, torrents, path, autoconnect=True):
        """Change torrents' location in the file system

        torrents: See `torrents` method
        path: Destination of the specified torrents
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents that were removed with the keys 'id'
                      and 'name'
            success: True if any torrents were found and had matching files,
                     False otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        if not os.path.isabs(path):
            # Transmission wants an absolute path, so we convert a relative
            # destination path starting from the default download path.
            response = await self._request(self.rpc.session_get, autoconnect=autoconnect)
            if not response.success:
                return Response(torrents=(), success=False, msgs=response.msgs)
            else:
                download_dir = response.result['download-dir']
                path = os.path.normpath(os.path.join(download_dir, path))

        def create_info_msg(t):
            if t['path'] != path:
                return (True, 'Moved to %s: %s' % (path, t['name']))
            else:
                return (False, 'Already in %s: %s' % (path, t['name']))

        return await self._torrent_action(self.rpc.torrent_set_location, torrents,
                                          keys=('name', 'path'),
                                          method_args={'move': True, 'location': path},
                                          check=create_info_msg,
                                          autoconnect=autoconnect)


    async def file_priority(self, priority, torrents, files, autoconnect=True):
        """Change download priority of individual torrent files

        priority: 'high', 'low', 'normal' or 'shun'
        torrents: See `torrents` method
        files: TorrentFileFilter object (or its string representation), sequence
               of (torrent ID, file ID) tuples or None for all files
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of matching Torrents that have matching files with
                      the keys 'id', 'name' and 'files'
            success: True if any torrents were found and had matching files,
                     False otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        response = await self.torrents(torrents, keys=('name', 'files'),
                                       autoconnect=autoconnect)
        if not response.success:
            return Response(torrents=(), success=False, msgs=response.msgs)
        else:
            torrents = ()
            torrent_ids = []
            msgs = []
            success = False

            if isinstance(files, str):
                files = TorrentFileFilter(files)

            # Set filter_files to a function that takes a TorrentFileTree and
            # returns a list of TorrentFiles.
            if files is None:
                filter_files = lambda ftree: tuple(ftree.files)
            elif isinstance(files, TorrentFileFilter):
                filter_files = lambda ftree: tuple(files.apply(ftree.files))
            elif isinstance(files, abc.Sequence) and \
                 all(isinstance(tid, int) and isinstance(fid, int) for tid,fid in files):
                filter_files = lambda ftree: tuple(f for f in ftree.files
                                                   if (f['tid'],f['id']) in files)
            else:
                raise ValueError("Invalid 'files' argument: {!r}".format(files))

            for t in sorted(response.torrents, key=lambda t: t['name'].lower()):
                # Filter torrent's files
                flist = filter_files(t['files'])
                if files is None:
                    msgs.append('{} file{}: {}'
                                .format(len(flist), '' if len(flist) == 1 else 's', t['name']))
                else:
                    if not flist:
                        msgs.append(ClientError('No matching files: {}'.format(t['name'])))
                    else:
                        msgs.append('{} matching file{}: {}'
                                    .format(len(flist), '' if len(flist) == 1 else 's', t['name']))
                success = len(flist) > 0 or success

                # Transmission wants a list of file indexes; luckily, the
                # file's ID is its index (see .torrent.TorrentFileTree).
                findexes = tuple(f['id'] for f in flist)
                if findexes:
                    response = await self._set_files_priority(priority, t['id'], findexes, autoconnect)
                    if response.success:
                        torrent_ids.append(t['id'])
                    msgs.extend(response.msgs)

        if torrent_ids:
            response = await self.torrents(torrent_ids, keys=('name', 'files'),
                                           autoconnect=autoconnect)
            if response.success:
                torrents = response.torrents
        return Response(torrents=torrents,
                        success=success,
                        msgs=msgs)

    async def _set_files_priority(self, priority, torrent_id, file_indexes, autoconnect):
        fi = tuple(file_indexes)
        if priority in ('high', 'normal', 'low'):
            return await self._torrent_action(
                self.rpc.torrent_set, (torrent_id,),
                method_args={'priority-%s' % priority: fi, 'files-wanted': fi},
                keys=(),
                autoconnect=autoconnect)
        elif priority == 'shun':
            return await self._torrent_action(
                self.rpc.torrent_set, (torrent_id,),
                method_args={'files-unwanted': fi},
                keys=(),
                autoconnect=autoconnect)
        else:
            raise ValueError('Invalid priority: {!r}'.format(priority))


    async def announce(self, torrents, autoconnect=True):
        """Re-announce to torrents' tracker

        torrents: See `torrents` method
        autoconnect: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of Torrents with the keys 'id' and 'name'
            success: True if any torrents were found, False otherwise
            msgs: list of strings/`ClientError`s caused by the request
        """
        from ..tkeys import Status
        import time
        def check(t):
            if t['status'] not in (Status.SEED, Status.LEECH):
                return (False, 'Not announcing inactive torrent: %s' % t['name'])
            elif t['timestamp-manual-announce-allowed'] > time.time():
                return (False, ('Not allowing manual announce until %s (in %s): %r' %
                                (t['timestamp-manual-announce-allowed'],
                                 t['timestamp-manual-announce-allowed'].delta, t['name'])))
            else:
                return (True, 'Announcing: %s' % t['name'])

        return await self._torrent_action(self.rpc.torrent_reannounce, torrents,
                                          check=check,
                                          keys=('name', 'status',
                                                'timestamp-manual-announce-allowed'),
                                          autoconnect=autoconnect)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Torrent commands for the TUI"""

from ...logging import make_logger
log = make_logger(__name__)

from shutil import get_terminal_size
TERMSIZE = get_terminal_size(fallback=(None, None))
import os

from ..base import torrent as base
from . import mixin
from .. import ExpectedResource
from ...utils import strwidth
from ...columns.tlist import COLUMNS as TLIST_COLUMNS
from ...columns.flist import COLUMNS as FLIST_COLUMNS
from ...columns.flist import create_directory_data
from ...columns.plist import COLUMNS as PLIST_COLUMNS


def _print_table(items, columns_wanted, COLUMN_SPECS):
    """Print table from a two-dimensional array of column objects

    `COLUMN_SPECS` maps column IDs to ColumnBase classes.  A column ID is any
    hashable object, but you probably want strings like 'name', 'id', 'date',
    etc.

    `columns_wanted` is a sequence of column IDs.

    `items` is a sequence of arbitrary objects that are used to create cell
    objects by passing them to the classes in `COLUMN_SPECS`.
    """

    # Create two-dimensional list to represent a table.  Each cell must be a
    # ColumnBase instance (see columns.tlist module).
    rows  = []
    for item in items:
        row = []
        for i,colname in enumerate(columns_wanted):
            cell = COLUMN_SPECS[colname](item)
            cell.index = i
            row.append(cell)
        rows.append(row)

    delimiter = '\t' if TERMSIZE.columns is None else '│'

    # Whether to print for a human or for a machine to read our output
    pretty_output = TERMSIZE.columns is not None

    def assemble_line(row):
        line = []
        for cell in row:
            if pretty_output:
                line.append(cell.get_string())
            else:
                line.append(str(cell.get_raw()))
        return delimiter.join(line)

    def assemble_headers():
        # This must be called after shrink_and_expand_to_fit() so we can
        # grab the final column widths from the first row.
        widths = tuple(cell.width for cell in rows[0])
        headers = []
        for colname,width in zip(columns_wanted, widths):
            header_items = COLUMN_SPECS[colname].header
            left  = header_items.get('left', '')
            right = header_items.get('right', '')
            space = ' '*(width - len(left) - len(right))
            header = ''.join((left, space, right))[:width]
            headers.append(header)
        return delimiter.join(headers)

    def shrink_and_expand_to_fit():
        log.debug('TTY width is %dx%d', TERMSIZE.columns, TERMSIZE.lines)

        def get_max_colwidth(colindex):
            # Return width of widest cell in all rows
            return max(strwidth(row[colindex].get_string())
                       for row in rows)

        def set_colwidth(colindex, width):
            # Set column width of all rows
            for row in rows:
                cell = row[colindex]
                cell.width = width

        def widest_columns():
            # List of columns sorted by width
            return sorted(range(len(columns_wanted)),
                          key=lambda colindex: get_max_colwidth(colindex),
                          reverse=True)

        # Expand column widths to make all cell values fit
        for colindex in range(len(columns_wanted)):
            colwidth = get_max_colwidth(colindex)
            set_colwidth(colindex, colwidth)

        # Rows should have identical column widths from now on, so we can
        # use the first row to check our progress.
        current_line = assemble_line(rows[0])
        current_width = strwidth(current_line)
        while current_width > TERMSIZE.columns:
            excess = current_width - TERMSIZE.columns
            widest = widest_columns()
            widest_0 = get_max_colwidth(widest[0])
            widest_1 = get_max_colwidth(widest[1])

            # Shorten widest column by difference to second widest column
            # (leaving them at the same width), but not by more than `excess`
            # characters and at least one character.

            # TODO: This is very slow when listing lots of rows in a small
            # terminal because the widest column is shrunk by only 1 character
            # before checking again.
            shorten_by = max(1, min(excess, widest_0 - widest_1))
            set_colwidth(widest[0], widest_0 - shorten_by)

            current_line = assemble_line(rows[0])
            current_width = strwidth(current_line)

    if rows:
        if not pretty_output:
            log.debug('Could not detect TTY size - assuming stdout is no TTY')
            headerstr = None
        elif TERMSIZE.columns < len(columns_wanted)*3:
            log.error('Terminal is too narrow for %d columns', len(columns_wanted))
            return False
        else:
            shrink_and_expand_to_fit()
            headerstr = '\033[1;4m' + assemble_headers() + '\033[0m'

        for linenum,row in enumerate(rows):
            if headerstr is not None and \
               linenum % (TERMSIZE.lines-1) == 0:
                log.info(headerstr)
            log.info(assemble_line(row))



class AddTorrentsCmd(base.AddTorrentsCmdbase,
                     mixin.make_request):
    provides = {'cli'}


class AnnounceTorrentsCmd(base.AnnounceTorrentsCmdbase,
                     mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class ListTorrentsCmd(base.ListTorrentsCmdbase,
                      mixin.make_request, mixin.select_torrents,
                      mixin.only_supported_columns):
    provides = {'cli'}
    srvapi = ExpectedResource  # TUI version of 'list' doesn't need srvapi
    async def make_tlist(self, tfilter, sort, columns):
        columns = self.only_supported_columns(columns, TLIST_COLUMNS)

        # Get wanted torrents and sort them
        if tfilter is None:
            keys = set(sort.needed_keys)
        else:
            keys = set(sort.needed_keys + tfilter.needed_keys)
        for colname in columns:
            keys.update(TLIST_COLUMNS[colname].needed_keys)
        response = await self.make_request(
            self.srvapi.torrent.torrents(tfilter, keys=keys),
            quiet=True)
        torrents = sort.apply(response.torrents)

        if torrents:
            _print_table(torrents, columns, TLIST_COLUMNS)
        return len(torrents) > 0


class ListFilesCmd(base.ListFilesCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.select_files,
                   mixin.only_supported_columns):
    provides = {'cli'}
    srvapi = ExpectedResource
    async def make_flist(self, tfilter, ffilter, columns):
        response = await self.make_request(
            self.srvapi.torrent.torrents(tfilter, keys=('name', 'files')),
            quiet=True)
        torrents = response.torrents

        if len(torrents) < 1:
            return False

        filelist = []
        for torrent in sorted(torrents, key=lambda t: t['name'].lower()):
            files, filtered_count = self._flatten_tree(torrent['files'], ffilter)
            filelist.extend(files)

        if filelist:
            columns = self.only_supported_columns(columns, FLIST_COLUMNS)
            _print_table(filelist, columns, FLIST_COLUMNS)
            return True
        else:
            if str(tfilter) != 'all':
                log.error('No matching files in {} torrents: {}'.format(tfilter, ffilter))
            else:
                log.error('No matching files: {}'.format(ffilter))
            return False

    def _flatten_tree(self, files, ffilter=None, _indent_level=0):
        """Return list of rows for _print_table

        `files` must be a nested mapping tree (i.e. TorrentFileTree).
        `ffilter` must be a TorrentFileFilter instance or None.
        """
        if TERMSIZE.columns is None:
            def indent_file_name(node):
                node['name'] = os.path.join(node['path'], node['name'])

            def indent_directory_name(node):
                node['name'] = node['path']
        else:
            def indent_file_name(node):
                node['name'] = '%s%s' % ('  '*(_indent_level), node['name'])
            indent_directory_name = indent_file_name

        flist = []
        filtered_count = 0
        for key,value in sorted(files.items(), key=lambda pair: pair[0].lower()):
            if value.nodetype == 'leaf':
                if ffilter is None or ffilter.match(value):
                    filenode = dict(value)  # Copy original TorrentFile
                    indent_file_name(filenode)
                    flist.append(filenode)
                else:
                    filtered_count += 1

            elif value.nodetype == 'parent':
                sub_flist, sub_filtered_count = self._flatten_tree(value, ffilter, _indent_level+1)
                dirnode = create_directory_data(key, value, sub_filtered_count)
                indent_directory_name(dirnode)
                flist.append(dirnode)
                flist.extend(sub_flist)

        return flist, filtered_count


class ListPeersCmd(base.ListPeersCmdbase,
                   mixin.make_request, mixin.select_torrents):
    provides = {'cli'}
    srvapi = ExpectedResource
    async def make_plist(self, tfilter, pfilter, sort, columns):
        response = await self.make_request(
            self.srvapi.torrent.torrents(tfilter, keys=('name', 'peers')),
            quiet=True)
        torrents = response.torrents

        if len(torrents) < 1:
            return False

        if pfilter is None:
            filter_peers = lambda peers: peers
        else:
            filter_peers = lambda peers: pfilter.apply(peers)

        peerlist = []
        for torrent in sorted(torrents, key=lambda t: t['name'].lower()):
            peerlist.extend(filter_peers(torrent['peers']))

        sort.apply(peerlist, inplace=True)

        if peerlist:
            _print_table(peerlist, columns, PLIST_COLUMNS)
            return True
        else:
            filter_is_relevant = lambda f: f and str(f) != 'all'

            if filter_is_relevant(pfilter):
                errmsg = 'No matching peers'
            else:
                errmsg = 'No peers'

            if filter_is_relevant(tfilter):
                errmsg += ' in {} torrents'.format(tfilter)

            if filter_is_relevant(pfilter):
                errmsg += ': {}'.format(pfilter)

            log.error(errmsg)
            return False


class MoveTorrentsCmd(base.MoveTorrentsCmdbase,
                      mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class PriorityCmd(base.PriorityCmdbase,
                  mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'cli'}


class RemoveTorrentsCmd(base.RemoveTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class StopTorrentsCmd(base.StopTorrentsCmdbase,
                      mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class StartTorrentsCmd(base.StartTorrentsCmdbase,
                       mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class VerifyTorrentsCmd(base.VerifyTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'cli'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for torrent list columns"""

from ..logging import make_logger
log = make_logger(__name__)

from . import ColumnBase


COLUMNS = {}


class TorrentName(ColumnBase):
    header = {'left': 'Torrent'}
    align = 'left'
    width = None
    may_have_wide_chars = True

    def get_value(self):
        return self.data['tname']

COLUMNS['name'] = TorrentName


class Client(ColumnBase):
    header = {'left': 'Client'}
    align = 'left'
    width = None

    def get_value(self):
        return self.data['client']

COLUMNS['client'] = Client


class Country(ColumnBase):
    header = {'left': 'Country'}
    align = 'right'
    width = 7

    def get_value(self):
        return self.data['country']

COLUMNS['country'] = Country


class IPAddress(ColumnBase):
    header = {'left': 'IP'}
    align = 'right'
    width = 15

    def get_value(self):
        return self.data['ip']

COLUMNS['ip'] = IPAddress


class Port(ColumnBase):
    header = {'left': 'Port'}
    align = 'right'
    width = 5

    def get_value(self):
        return self.data['port']

COLUMNS['port'] = Port


class Progress(ColumnBase):
    header = {'right': '%'}
    width = 5

    def get_value(self):
        return self.data['progress']

COLUMNS['progress'] = Progress


class RateDown(ColumnBase):
    header = {'left': 'Dn', 'right': '?/s'}
    width = 6

    def get_value(self):
        return self.data['rate-down']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-down'] = RateDown


class RateUp(ColumnBase):
    header = {'left': 'Up', 'right': '?/s'}
    width = 6

    def get_value(self):
        return self.data['rate-up']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-up'] = RateUp


class ETA(ColumnBase):
    header = {'left': 'ETA'}
    width = 5

    def get_value(self):
        return self.data['eta']

    def get_raw(self):
        return int(self.get_value())

COLUMNS['eta'] = ETA


class EstimatedPeerRate(ColumnBase):
    header = {'left': 'Est', 'right': '?/s'}
    width = 7

    def get_value(self):
        return self.data['rate-est']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-est'] = EstimatedPeerRate



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import urwid
import re
import operator
import os

from . import keymap as km

class CLIEditWidget(urwid.Edit):
    """Edit widget with readline keybindings callbacks and a history"""

    def __init__(self, *args, on_change=None, on_accept=None, on_cancel=None,
                 history_file=None, **kwargs):
        kwargs['align'] = kwargs['align'] if 'align' in kwargs else 'left'
        kwargs['wrap'] = kwargs['wrap'] if 'wrap' in kwargs else 'clip'
        self._on_change = on_change
        self._on_accept = on_accept
        self._on_cancel = on_cancel
        self._edit_text_cache = ''
        self._history = []
        self._history_pos = -1
        self.history_file = history_file
        return super().__init__(*args, **kwargs)

    def keypress(self, size, key):
        size = (size[0],)
        text_before = self.get_edit_text()
        if self._command_map[key] is km.CURSOR_UP:
            self._set_history_prev()
            key = None
        elif self._command_map[key] is km.CURSOR_DOWN:
            self._set_history_next()
            key = None
        elif self._command_map[key] is km.DELETE_TO_EOL:
            self.edit_text = self.edit_text[:self.edit_pos]
            key = None
        elif self._command_map[key] is km.DELETE_LINE:
            self.set_edit_text('')
            key = None
        elif self._command_map[key] is km.DELETE_CHAR_UNDER_CURSOR:
            return super().keypress(size, 'delete')
        elif self._command_map[key] is km.CURSOR_WORD_RIGHT:
            self.move_to_next_word(forward=True)
            key = None
        elif self._command_map[key] is km.CURSOR_WORD_LEFT:
            self.move_to_next_word(forward=False)
            key = None
        elif self._command_map[key] is km.DELETE_WORD_LEFT:
            start_pos = self.edit_pos
            end_pos = self.move_to_next_word(forward=True)
            if end_pos != None:
                self.set_edit_text(self.edit_text[:start_pos] + self.edit_text[end_pos:])
            self.edit_pos = start_pos
            key = None
        elif self._command_map[key] is km.DELETE_WORD_RIGHT:
            end_pos = self.edit_pos
            start_pos = self.move_to_next_word(forward=False)
            if start_pos != None:
                self.set_edit_text(self.edit_text[:start_pos] + self.edit_text[end_pos:])
            key = None
        elif self._on_accept is not None and self._command_map[key] is km.ACTIVATE:
            self._append_to_history(self.edit_text)
            self._on_accept(self)
            self._reset()
            key = None
        elif self._on_cancel is not None and self._command_map[key] is km.CANCEL:
            self._on_cancel(self)
            self._reset()
            key = None
        elif key == 'space':
            return super().keypress(size, ' ')
        else:
            key = super().keypress(size, key)
        text_after = self.get_edit_text()
        if self._on_change is not None and text_before != text_after:
            self._on_change(self)
        return key

    def move_to_next_word(self, forward=True):
        if forward:
            match_iterator  = re.finditer(r'(\b\W+|$)', self.edit_text,
                                          flags=re.UNICODE)
            match_positions = [m.start() for m in match_iterator]
            op = operator.gt
        else:
            match_iterator  = re.finditer(r'(\w+\b|^)', self.edit_text,
                                          flags=re.UNICODE)
            match_positions = reversed([m.start() for m in match_iterator])
            op = operator.lt
        for pos in match_positions:
            if op(pos, self.edit_pos):
                self.set_edit_pos(pos)
                return pos

    def _reset(self):
        self._history_pos = -1

    def _set_history_prev(self):
        # Remember whatever is currently in line when user starts exploring history
        if self._history_pos == -1:
            self._edit_text_cache = self.edit_text
        if self._history_pos+1 < len(self._history):
            self._history_pos += 1
            self.edit_text = self._history[self._history_pos]

    def _set_history_next(self):
        # The most recent history entry is at index 0; -1 is the current line
        # that is not yet in history.
        if self._history_pos > -1:
            self._history_pos -= 1
        if self._history_pos == -1:
            self.edit_text = self._edit_text_cache  # Restore current line
        else:
            self.edit_text = self._history[self._history_pos]

    def _read_history(self):
        if self._history_file:
            self._history = _read_lines(self._history_file)

    def _append_to_history(self, line):
        # Don't add the same line twice in a row
        if self._history and self._history[0] == line:
            return
        self._history.insert(0, line)
        if self._history_file is not None:
            _append_line(self._history_file, line)

    @property
    def history_file(self):
        return self._history_file

    @history_file.setter
    def history_file(self, path):
        if path is None:
            self._history_file = None
        else:
            if _mkdir(os.path.dirname(path)) and _testwrite(path):
                self._history_file = os.path.abspath(os.path.expanduser(path))
                self._read_history()


def _read_lines(filepath):
    if  os.path.exists(filepath):
        try:
            with open(filepath, 'r') as f:
                return [line.strip() for line in reversed(f.readlines())]
        except OSError as e:
            log.error("Can't read history file {}: {}"
                      .format(filepath, e.strerror))
    return []

def _mkdir(path):
    try:
        if path and not os.path.exists(path):
            os.makedirs(path)
    except OSError as e:
        log.error("Can't create directory {}: {}".format(path, e.strerror))
    else:
        return True

def _testwrite(path):
    try:
        open(path, 'a')
    except OSError as e:
        log.error("Can't append to history file {}: {}".format(path, e.strerror))
    else:
        return True

def _append_line(filepath, line):
    try:
        with open(filepath, 'a') as f:
            f.write(line.strip() + '\n')
    except OSError as e:
        log.error("Can't append to history file {}: {}".format(filepath, e.strerror))
    else:
        return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid

from .group import Group


class Table():
    """Manage rows, columns and headers as in a table

    `columns`: Dictionary that maps column IDs to column classes

    A column class is any class that creates a cell widget when called without
    arguments and has the attributes 'header' and 'width'.

        - 'header' is a cell widget that is displayed at the top of the
          column.

        - 'width' is passed on to the 'options' argument of Group.add()

    By setting the `columns` property to column IDs, columns are displayed or
    hidden in each existing or newly added row.
    """
    def __init__(self, **columns):
        self._colspecs = columns
        self._enabled_columns = []
        self._headers = Group(cls=urwid.Columns, dividechars=1)
        self._members = {}
        self.columns = columns

    def register(self, member_id):
        """Add a new row

        Create a new Group(cls=Columns) object and fill it with enabled column
        cells which can then be retrieved with `get_row(member_id)`.
        """
        member = Group(cls=urwid.Columns, dividechars=1)
        for colname in self._enabled_columns:
            cellcls = self._colspecs[colname]
            cellwidget = cellcls()
            member.add(colname, cellwidget, options=cellcls.width, removable=True)
        self._members[member_id] = member

    def get_row(self, member_id):
        """Return a row, i.e. a Group(cls=Columns) object created by register()"""
        return self._members[member_id]

    @property
    def headers(self):
        """Header row (a Group(cls=Columns) object)"""
        return self._headers

    @property
    def columns(self):
        """Currently enabled/displayed column IDs"""
        return self._enabled_columns

    @columns.setter
    def columns(self, columns):
        for col in columns:
            if col not in self._colspecs:
                raise ValueError('Unknown column name: {!r}'.format(col))

        # Remove all columns
        self._headers.clear()
        for member in self._members.values():
            member.clear()
        self._enabled_columns = []

        # Add wanted columns
        for colname in columns:
            # Add header
            cellcls = self._colspecs[colname]
            self._headers.add(colname, cellcls.header, options=cellcls.width, removable=True)
            self._enabled_columns.append(colname)

            # Add new column to all members
            for member in self._members.values():
                cellwidget = cellcls()
                member.add(colname, cellwidget, options=cellcls.width, removable=True)

    def clear(self):
        """Remove all registered rows"""
        self._members = {}


class ColumnHeaderWidget(urwid.WidgetWrap):
    """Column widget with left and right text"""

    def __init__(self, left='', right=''):
        self._left = urwid.Text(left, wrap='clip')
        self._right = urwid.Text(right, wrap='clip')
        widget = urwid.Columns([self._left, ('pack', self._right)],
                               dividechars=0)
        super().__init__(widget)

    @property
    def left(self):
        """String on the left side"""
        return self._left.text

    @left.setter
    def left(self, string):
        self._left.set_text(string)

    @property
    def right(self):
        """String on the right side"""
        return self._right.text

    @right.setter
    def right(self, string):
        self._right.set_text(string)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import urwid
import urwidtrees
from collections import abc
import builtins

from .. import main as tui
from ..table import Table
from .flist_columns import TUICOLUMNS
from ...columns.flist import (create_directory_data, create_directory_name)

COLUMNS_FOCUS_MAP = {}
for col in TUICOLUMNS.values():
    COLUMNS_FOCUS_MAP.update(col.style.focus_map)


class FileWidget(urwid.WidgetWrap):
    def __init__(self, tfile, row):
        self._cells = row
        super().__init__(urwid.AttrMap(row, 'filelist', 'filelist.focused'))
        self.update(tfile)

    def update(self, tfile):
        for widget in self._cells.widgets:
            widget.update(tfile)
        self._tfile = tfile

    @property
    def torrent_id(self):
        return self._tfile['tid']

    @property
    def file_id(self):
        return self._tfile['id']

    @property
    def is_marked(self):
        return self._cells.marked.is_marked

    @is_marked.setter
    def is_marked(self, is_marked):
        self._cells.marked.is_marked = bool(is_marked)

    @property
    def nodetype(self):
        """'parent' or 'leaf'"""
        return self._tfile.nodetype

    def __repr__(self):
        return '<{} {!r}>'.format(type(self).__name__, self._tfile['name'])


from urwidtrees.decoration import ArrowTree
class FileTreeDecorator(ArrowTree):
    """urwidtrees decorator for TorrentFiles and TorrentFileTrees"""

    def __init__(self, torrents, keymap, table, ffilter):
        self._filewidgetcls = keymap.wrap(FileWidget, context='file')
        self._table = table
        self._ffilter = ffilter
        self._widgets = {}
        self._filtered_counts = {}
        forest = self._create_file_forest(torrents)
        super().__init__(forest, indent=2)

    def _file_is_filtered(self, tfile):
        if self._ffilter is None:
            return False  # No filter specified
        elif isinstance(self._ffilter, (abc.Sequence, abc.Set)):
            # ffilter is a collection of file IDs
            return not tfile['id'] in self._ffilter
        else:
            # ffilter is a TorrentFileFilter instance
            return not self._ffilter.match(tfile)

    def _create_file_forest(self, torrents):
        # Create a list of nested trees in SimpleTree format.  But the leaves
        # are mappings instead of widgets.  Each mapping contains the
        # information that the `decorate` method needs to create a widget.

        def create_tree(node, content):
            if content.nodetype == 'leaf':
                # Torrent has a single file and no directories
                if not self._file_is_filtered(content):
                    return (content, None)
                else:
                    return None

            elif content.nodetype == 'parent':
                # Torrent has at least one directory
                tree = []
                filtered_count = 0
                for k,v in sorted(content.items(), key=lambda pair: pair[0].lower()):
                    if v.nodetype == 'leaf':
                        if not self._file_is_filtered(v):
                            tree.append((v, None))
                            nonlocal fcount
                            fcount += 1
                        else:
                            filtered_count += 1
                    elif v.nodetype == 'parent':
                        dirnode = create_directory_data(name=k, tree=v)
                        tree.append(create_tree(dirnode, v))

                node_id = (node['tid'], node['id'])
                self._filtered_counts[node_id] = filtered_count
                node['name'] = create_directory_name(node['name'], filtered_count)
                return (node, tree or None)

        fcount = 0
        forest = []  # Multiple trees as siblings
        for t in sorted(torrents, key=lambda t: t['name'].lower()):
            filetree = t['files']
            # This works because t['files'] always has 1 item: the torrent's name
            rootnodename = next(iter(filetree.keys()))
            rootnode = create_directory_data(rootnodename, tree=filetree)
            tree = create_tree(rootnode, filetree[rootnodename])
            if tree is not None:
                forest.append(tree)
        self.filecount = fcount
        return forest

    def decorate(self, pos, data, is_first=True):
        # We can use the tree position as table ID
        self._table.register(pos)
        row = self._table.get_row(pos)

        # We use parent's decorate() method to give the name column a tree
        # structure.  But we also need the original update() method so we can
        # apply new data to the widget.  This is dirty but it works.
        update_method = row.name.update
        decowidget = super().decorate(pos, row.name, is_first=is_first)
        decowidget.update = update_method
        row.replace('name', decowidget)

        # Wrap the whole row in a FileWidget with keymapping.  This also
        # applies all the other values besides the name (size, progress, etc).
        file_widget = self._filewidgetcls(data, row)
        node_id = (data['tid'], data['id'])
        self._widgets[node_id] = file_widget
        return urwid.AttrMap(file_widget, attr_map=None, focus_map=COLUMNS_FOCUS_MAP)

    def update(self, torrents):
        widgets = self._widgets
        for t in torrents:
            tid = t['id']

            # Update file nodes
            for f in t['files'].files:
                fid = f['id']
                node_id = (tid, fid)
                if node_id in widgets:
                    widgets[node_id].update(f)

            # Update directory nodes
            for name,content in t['files'].folders:
                fids = frozenset(f['id'] for f in content.files)
                node_id = (tid, fids)
                if node_id in widgets:
                    filtered_count = self._filtered_counts[node_id]
                    data = create_directory_data(name, tree=content,
                                                 filtered_count=filtered_count)
                    widgets[node_id].update(data)

    @property
    def widgets(self):
        """Yield all file and directory widgets in this tree"""
        yield from self._widgets.values()


class FileListWidget(urwid.WidgetWrap):
    def __init__(self, srvapi, tfilter, ffilter, columns, title=None):
        self._ffilter = ffilter
        self._torrents = ()
        self._marked = set()
        self._initialized = False

        # Create the fixed part of the title (everything minus the number of files listed)
        # If title is not given, create one from filters
        if title is None:
            tfilter_str = str(tfilter or 'all')
            if ffilter is None:
                title = tfilter_str
            else:
                title = '%s files of %s torrents' % (ffilter, tfilter_str)
        self._title_base = title
        self.title_updater = None

        self._table = Table(**TUICOLUMNS)
        self._table.columns = columns

        self._listbox = urwid.ListBox(urwid.SimpleListWalker([]))
        pile = urwid.Pile([
            ('pack', urwid.AttrMap(self._table.headers, 'filelist.header')),
            self._listbox
        ])
        super().__init__(urwid.AttrMap(pile, 'filelist'))

        self._poller = srvapi.create_poller(
            srvapi.torrent.torrents, tfilter, keys=('files', 'name'))
        self._poller.on_response(self._handle_response)

    def _handle_response(self, response):
        if response is None or not response.torrents:
            self.clear()
        else:
            if self._initialized:
                self._update_listitems(response.torrents)
            else:
                self._init_listitems(response.torrents)
        if self.title_updater is not None:
            self.title_updater(self.title)
        self._invalidate()

    def _init_listitems(self, torrents):
        self.clear()
        if torrents:
            self._filetree = FileTreeDecorator(torrents, tui.keymap,
                                               self._table, self._ffilter)
            self._listbox.body = urwidtrees.widgets.TreeListWalker(self._filetree)
            self._listbox._invalidate()
            self._initialized = True

    def _update_listitems(self, torrents):
        if torrents:
            self._filetree.update(torrents)

    def clear(self):
        """Remove all list items"""
        self._table.clear()
        self._listbox.body = urwid.SimpleListWalker([])
        self._listbox._invalidate()
        self._initialized = False

    @property
    def title(self):
        if hasattr(self, '_filetree'):
            return '%s [%d]' % (self._title_base, self._filetree.filecount)
        else:
            return '%s [0]' % self._title_base

    def update(self):
        """Call `clear` and then poll immediately"""
        self.clear()
        self._poller.poll()

    @property
    def focus_position(self):
        positions = tuple(self._filetree.positions())
        return positions.index(self._listbox.focus_position)

    @focus_position.setter
    def focus_position(self, focus_position):
        positions = tuple(self._filetree.positions())
        i = min(focus_position, len(positions)-1)
        try:
            self._listbox.focus_position = positions[i]
        except KeyError:
            pass

    @property
    def focused_file(self):
        """Focused FileWidget instance"""
        focused = self._listbox.focus
        if focused is not None:
            return focused.original_widget

    @property
    def focused_torrent_id(self):
        """Torrent ID of the focused file's torrent"""
        focused = self.focused_file
        if focused is not None:
            return focused.torrent_id

    @property
    def focused_file_ids(self):
        """File IDs of the focused files in a tuple"""
        focused = self.focused_file
        if focused is not None:
            # The focused widget in the list can be a file or a directory.  If
            # it's a directory, the 'file_id' property returns the IDs of all
            # the contained files recursively.
            fid = focused.file_id
            return tuple(fid) if isinstance(fid, (abc.Sequence, abc.Set)) else (fid,)

    def all_children(self, pos):
        """Yield (position, widget) tuples of all sub-nodes (leaves and parents)"""
        ft = self._filetree
        lb = self._listbox
        def recurse(subpos):
            widget = lb.contents[subpos][0].original_widget
            if ft.is_leaf(subpos):
                yield (subpos, widget)
            else:
                # Yield sub-parent nodes, but not the starting node that was
                # passed to us
                if subpos != pos:
                    yield (subpos, widget)

                new_subpos = ft.first_child_position(subpos)
                while new_subpos is not None:
                    yield from recurse(new_subpos)
                    new_subpos = ft.next_sibling_position(new_subpos)

        yield from recurse(pos)

    def mark(self, toggle=False, all=False):
        """Mark the currently focused item or all items"""
        self._set_mark(True, toggle=toggle, all=all)

    def unmark(self, toggle=False, all=False):
        """Unmark the currently focused item or all items"""
        self._set_mark(False, toggle=toggle, all=all)

    @property
    def marked(self):
        """Generator that yields FileWidgets"""
        yield from self._marked

    def _set_mark(self, mark, toggle=False, all=False):
        if toggle:
            focused = self.focused_file
            if focused is not None:
                mark = not focused.is_marked

        def get_widget(pos):
            return self._listbox.contents[pos][0].original_widget

        def mark_leaves(pos, mark):
            get_widget(pos).is_marked = mark

            for subpos,widget in self.all_children(pos):
                if widget.nodetype == 'leaf':
                    widget.is_marked = mark
                    if mark:
                        self._marked.add(widget)
                    else:
                        self._marked.discard(widget)

                elif widget.nodetype == 'parent':
                    mark_leaves(subpos, mark)

        # First we mark all leaves recursively.
        if all:
            # Top ancestor node positions are (0,), (1,), (3,) etc
            for pos in self._filetree.positions():
                if len(pos) == 1:
                    mark_leaves(pos, mark)
        else:
            mark_leaves(self._listbox.focus_position, mark)
        assert builtins.all(m.nodetype == 'leaf' for m in self._marked)

        # A parent node is marked only if all its children are marked.  To check
        # that, we walk through every ancestor up to the top and check all its
        # children.  There is no need to check the children of other parent
        # nodes (uncles, great uncles, etc) because they should already be
        # marked properly from previous runs.

        def all_children_marked(pos):
            marked = True
            childpos = self._filetree.first_child_position(pos)
            while childpos is not None:
                marked = marked and get_widget(childpos).is_marked
                childpos = self._filetree.next_sibling_position(childpos)
            return marked

        parpos = self._filetree.parent_position(self._listbox.focus_position)
        while parpos is not None:
            parwidget = get_widget(parpos)
            parwidget.is_marked = all_children_marked(parpos)
            parpos = self._filetree.parent_position(parpos)

    def refresh_marks(self):
        """Redraw the "marked" column in all rows"""
        for widget in self._filetree.widgets:
            widget.is_marked = widget.is_marked



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import logging
import urwid
import time
import asyncio


class UILogRecordHandler(logging.Handler):
    """Feed LogRecords to a LogWidget"""

    def __init__(self, logwidget):
        super().__init__()
        self._logwidget = logwidget

    def emit(self, record):
        self._logwidget.add(record)


class LogWidget(urwid.WidgetWrap):
    """Present LogRecords from logging module in a ListBox Widget"""

    def __init__(self, maxrows=10, autohide_delay=0, loop=None):
        self.maxrows = maxrows
        self.autohide_delay = autohide_delay
        self._autohide_handle = None
        self._loop = loop if loop is not None else asyncio.get_event_loop()
        self._listbox = urwid.ListBox(urwid.SimpleListWalker([]))
        super().__init__(urwid.AttrMap(self._listbox, 'log'))

        self._root_logger = logging.getLogger()
        self._orig_handlers = []
        self._handler = UILogRecordHandler(self)

        # Copy formatter first handler (there should be only one formatter anyway)
        handlers = self._root_logger.handlers
        self._handler.setFormatter(handlers[0].formatter)

        # Don't log debugging messages if we're already logging to a file
        for h in handlers:
            if isinstance(h, logging.FileHandler):
                self._handler.setLevel(logging.INFO)

    def enable(self):
        # Store any stdout/stderr handlers
        for h in tuple(self._root_logger.handlers):
            if h.stream.name in ('<stderr>', '<stdout>'):
                self._orig_handlers.append(h)
                self._root_logger.removeHandler(h)
        self._root_logger.addHandler(self._handler)
        log.debug('Started logging to UI')

    def disable(self):
        # Restore any previously stored handlers
        for h in tuple(self._orig_handlers):
            self._root_logger.addHandler(h)
            self._orig_handlers.remove(h)

        try:
            self._root_logger.handlers.remove(self._handler)
        except ValueError:
            pass
        else:
            log.debug('Stopped logging to UI')

        if self._autohide_handle is not None:
            self._autohide_handle.cancel()

    def add(self, record):
        msg = ''
        if record.levelno >= logging.WARNING:
            style = 'error'
        elif record.levelno > logging.DEBUG:
            style = 'info'
        else:
            style = 'debug'
            msg = '[{}] '.format(record.name)
        msg += record.getMessage()

        # Indicate identical messages instead of spamming the log
        if len(self._listbox.body) > 0 and self._listbox.body[-1].text == msg:
            self._listbox.body[-1].dupes += 1
        else:
            self._listbox.body.append(LogEntry(msg, style))
            self._listbox.focus_position = len(self._listbox.body)-1  # Scroll to end of log
        self._invalidate_rows()
        self._maybe_show_temporarily()

    # TODO: The autohide functionality shouldn't be in this widget.
    def _maybe_show_temporarily(self):
        """Show log widget if hidden and hide it again after delay"""
        from .main import widgets
        if self.autohide_delay > 0 and \
           (not widgets.visible('log') or self._autohide_handle is not None):
            if self._autohide_handle is not None:
                self._autohide_handle.cancel()
            widgets.show('log')
            def hide():
                widgets.hide('log')
                self._autohide_handle = None
            self._autohide_handle = self._loop.call_later(self.autohide_delay, hide)

    def rows(self, size, focus=False):
        if hasattr(self, '_rows'):
            return self._rows
        else:
            (maxcols, maxrows) = (size[0], self.maxrows)
            walker = self._listbox.body
            rows = 0

            if len(walker) >= maxrows:
                rows = maxrows
            else:
                for widget in walker:
                    rows += widget.rows(size=(maxcols,))
                    if rows >= maxrows:
                        break
            self._rows = min(rows, maxrows)
            return self._rows

    def render(self, size, focus=False):
        size = (size[0], self.rows(size, focus))
        return super().render(size, focus)

    def _invalidate_rows(self):
        try:
            delattr(self, '_rows')
        except AttributeError:
            pass

    def clear(self):
        self._listbox.body[:] = []
        self._invalidate_rows()

    @property
    def entries(self):
        yield from self._listbox.body

    def selectable(self):
        return False


class LogEntry(urwid.WidgetWrap):
    @staticmethod
    def _make_timestamp():
        return time.strftime('%H:%M:%S')

    def __init__(self, message, style):
        self._dupes = 0
        self._widgets = {
            'timestamp':    urwid.Text(self._make_timestamp()),
            'dupes':        urwid.Text(''),
            'dupes_spacer': urwid.Text(''),
            'message':      urwid.Text(str(message)),
        }
        super().__init__(urwid.Columns([
            ('pack', urwid.AttrMap(self._widgets['timestamp'],    'log.timestamp')),
            ('pack', urwid.AttrMap(urwid.Text(' '),               'log')),
            ('pack', urwid.AttrMap(self._widgets['dupes'],        'log.dupecount')),
            ('pack', urwid.AttrMap(self._widgets['dupes_spacer'], 'log')),
            urwid.AttrMap(self._widgets['message'], 'log.'+style),
        ], dividechars=0))

    @property
    def text(self):
        return self._widgets['message'].text

    @property
    def dupes(self):
        return self._dupes

    @dupes.setter
    def dupes(self, dupes):
        self._dupes = dupes
        self._widgets['timestamp'].set_text(self._make_timestamp())
        if dupes > 0:
            self._widgets['dupes'].set_text('x' + str(dupes+1))
            self._widgets['dupes_spacer'].set_text(' ')
        else:
            self._widgets['dupes'].set_text('')
            self._widgets['dupes_spacer'].set_text('')

    def __repr__(self):
        return '<{} {!r}>'.format(type(self).__name__, self._widgets['message'].text)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Indent and expand tabs and wrap lines"""

from textwrap import wrap
from shutil import get_terminal_size

def _explode(lines, indent):
    """Split each line in `lines` at tabstops

    Tabstops at the beginning of each line are expanded to 2 spaces.

    Return list of rows (a row being a list of strings, i.e. cells)
    """
    exploded = []
    for line in lines:
        indentions = 0
        while line.startswith('\t'):
            indentions += 1
            line = line[1:]
        splitline = line.split('\t')
        for i in range(indentions):
            splitline.insert(0, ' '*indent)
        exploded.append(splitline)
    return exploded

def _split_sections(lines):
    """Split `lines` into sections of lines that have the same amount of parts"""
    sections = []
    while lines:
        line = lines.pop(0)
        partnum = len(line)
        section = [line]
        while lines and len(lines[0]) == partnum:
            line = lines.pop(0)
            section.append(line)
        sections.append(section)
    return sections

def _col_widths(lines):
    """Return a list of maximum column widths"""
    widths = []
    for line in lines:
        for i,part in enumerate(line):
            while len(widths) <= i:
                widths.append(0)
            widths[i] = max(len(part), widths[i])
    return widths

def _join_line(parts, widths, maxwidth):
    """Expand and wrap all parts of `line`"""
    line = ''
    for colwidth,part in zip(widths, parts):
        part = part.ljust(colwidth)
        if len(line) + len(part) > maxwidth:
            line = _wrapindent(part, prefix=line, width=maxwidth)
        else:
            line += part
    return line

def _wrapindent(text, prefix, width):
    """Prepend `prefix` to text and wrap the result if longer than `width`"""
    indent = ' '*len(prefix)
    parts = wrap(prefix+text, width=width, subsequent_indent=indent)
    return '\n'.join(parts)

def expand(lines, indent=4, maxwidth=100):
    """Expand all tabstops (\t) in each line intelligently

    "Intelligently" means that consecutive lines with the same amount of '\t'
    characters are treated like a table, giving each column the same space.

    Return `lines` with all tabs expanded
    """
    width = min(get_terminal_size()[0], maxwidth)
    expanded = []
    for section in _split_sections(_explode(lines, indent)):
        widths = _col_widths(section)
        for line in section:
            expanded.append(_join_line(line, widths, width))
    return expanded



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import os
from collections import abc
from blinker import Signal


class Settings():
    """Dict of ordered Value objects"""

    def __init__(self, *values):
        # _values and _values_dict hold the same instances; the list is to
        # preserve order and the dict is for access via __getitem__, etc.
        self._values = []
        self._values_dict = {}
        self._on_change = Signal()
        self.load(*values)

    def add(self, value):
        self._values.append(value)
        self._values_dict[value.name] = value

    def load(self, *values):
        for v in values:
            self.add(v)

    @property
    def names(self):
        yield from self._values_dict.keys()

    @property
    def values(self):
        yield from self._values

    def on_change(self, callback, autoremove=True):
        """Run `callback` every time a value changes with the value

        If `autoremove` is True, stop calling callback once it is garbage
        collected.
        """
        self._on_change.connect(callback, weak=autoremove)

    def __getitem__(self, name):
        return self._values_dict[name]

    def __setitem__(self, name, value):
        self._values_dict[name].set(value)
        self._on_change.send(self._values_dict[name])

    def __contains__(self, name):
        return name in self._values_dict


class ValueBase():
    """Name:value pair with validation, default value and description"""

    def __init__(self, name, default, description='No description available'):
        """Create new value"""
        self.__name = name
        self.__value = self.__default = self.convert(default)
        self.__description = description
        self.__on_change = Signal()

    # Must be set by derived classes
    type = NotImplemented
    typename = '<NOT IMPLEMENTED>'

    @property
    def value(self): return self.__value

    @property
    def name(self): return self.__name

    @property
    def default(self): return self.__default

    @property
    def description(self): return self.__description

    def on_change(self, callback, autoremove=True):
        """Pass this object to `callback` every time its value changes

        `callback` may raise ValueError to revert the change (see `set`).

        If `autoremove` is True, stop calling callback once it is garbage
        collected.
        """
        self.__on_change.connect(callback, weak=autoremove)

    def set(self, value):
        """Change value if valid, reset to default if None

        Callbacks connected to `on_change` are passed this object every time a
        value is changed.  If a callback raises ValueError, the change is
        reverted and a ValueError is raised.
        """
        if value is None:
            value = self.__default
        try:
            new_value = self.convert(value)
            self.validate(new_value)
        except ValueError as e:
            raise ValueError('{} = {}: {}'.format(self.name, self.str(value), e))
        else:
            prev_value = self.__value
            self.__value = new_value
            # Callbacks can revert the change by raising ValueError
            try:
                self.__on_change.send(self)
            except ValueError as e:
                self.__value = prev_value
                raise ValueError('{} = {}: {}'.format(self.name, self.str(value), e))

    def get(self):
        """Return current value"""
        return self.value

    def validate(self, value):
        """Raise ValueError if value is not valid"""
        if not isinstance(value, self.type):
            raise ValueError('Not a {}'.format(self.typename))

    def convert(self, value):
        """Try to convert value to correct type before validation (e.g. str->int)

        Raise ValueError if impossible"""
        if isinstance(value, self.type):
            return value
        try:
            if isinstance(value, abc.Iterable):
                return self.type(''.join(value))
            else:
                return self.type(value)
        except Exception:
            raise ValueError('Not a {}'.format(self.typename))

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.value == other.value
        else:
            return self.value == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def str(self, value=None, default=False):
        """Return prettily stringified value

        value: None to return current value, or specific value
        default: Whether to return current or default value
        """
        if default:
            text = str(self.default)
        elif value is not None:
            text = str(value)
        else:
            text = str(self.value)

        if text and (text[0] == ' ' or text[-1] == ' '):
            return repr(text)
        else:
            return text

    def __str__(self):
        return self.str()

    def __repr__(self):
        return '{}={!r}'.format(self.name, self.value)


class StringValue(ValueBase):
    type = str
    @property
    def typename(self):
        text = 'string'
        if ((self._minlen == 1 or self._minlen == None) and
            (self._maxlen == 1 or self._maxlen == None)):
            chrstr = 'character'
        else:
            chrstr = 'characters'
        if self._minlen is not None and self._maxlen is not None:
            if self._minlen == self._maxlen:
                text += ' of {} {}'.format(self._minlen, chrstr)
            else:
                text += ' of {} to {} {}'.format(self._minlen, self._maxlen, chrstr)
        elif self._minlen is not None:
            text += ' of at least {} {}'.format(self._minlen, chrstr)
        elif self._maxlen is not None:
            text += ' of at most {} {}'.format(self._maxlen, chrstr)
        return text

    def __init__(self, *args, minlen=None, maxlen=None, **kwargs):
        if minlen is not None and maxlen is not None:
            assert minlen <= maxlen, 'minimum string length must be <= maximum'
        assert minlen is None or minlen >= 0, 'minimum string length must be >= 0'
        assert maxlen is None or maxlen >= 0, 'maximum string length must be >= 0'
        self._minlen = minlen
        self._maxlen = maxlen
        super().__init__(*args, **kwargs)

    def validate(self, string):
        if self._maxlen is not None and len(string) > self._maxlen:
            raise ValueError('Too long (maximum length is {})'.format(self._maxlen))
        if self._minlen is not None and len(string) < self._minlen:
            raise ValueError('Too short (minimum length is {})'.format(self._minlen))


class PathValue(StringValue):
    typename = 'path'

    def __init__(self, *args, mustexist=False, **kwargs):
        self.mustexist = mustexist
        super().__init__(*args, **kwargs)

    def convert(self, path):
        return os.path.expanduser(path)

    def validate(self, path):
        if self.mustexist and not os.path.exists(path):
            raise ValueError('No such file or directory')

    def str(self, value=None, default=False):
        """Tildify string"""
        path = super().str(value=value, default=default)
        if path.startswith(os.environ['HOME']):
            path = '~' + path[len(os.environ['HOME']):]
        return path


class NumberValue(ValueBase):
    type = float
    _numbertype = 'rational'
    valuesyntax = '[+=|-=]<NUMBER>'

    def __init__(self, *args, min=None, max=None, **kwargs):
        if min is not None and max is not None:
            assert min <= max, 'minimum must be smaller or equal than maximum'
        self._min = min
        self._max = max
        super().__init__(*args, **kwargs)

    @property
    def typename(self):
        text = '{} number'.format(self._numbertype)
        if self._min is not None and self._max is not None:
            text += ' {}-{}'.format(self._min, self._max)
        elif self._min is not None:
            text += ' >= {}'.format(self._min)
        elif self._max is not None:
            text += ' <= {}'.format(self._max)
        return text

    def convert(self, value):
        if isinstance(value, str) and len(value) >= 3:
            if value[0:2] == '+=':
                return self.value + super().convert(value[2:].strip())
            elif value[0:2] == '-=':
                return self.value - super().convert(value[2:].strip())
        return super().convert(value)

    def validate(self, value):
        super().validate(value)
        if self._min is not None and value < self._min:
            raise ValueError('Too small (minimum is {})'.format(self._min))
        elif self._max is not None and value > self._max:
            raise ValueError('Too big (maximum is {})'.format(self._max))

class IntegerValue(NumberValue):
    type = int
    _numbertype = 'integer'


TRUE = ('enable', 'yes', 'on', 'true', '1')
FALSE = ('disable', 'no', 'off', 'false', '0')
class BooleanValue(ValueBase):
    type = bool
    typename = 'boolean'
    valuesyntax = '[%s]' % '|'.join('/'.join((t,f)) for t,f in zip(TRUE, FALSE))

    def convert(self, value):
        if isinstance(value, bool):
            return value
        elif isinstance(value, int):
            if value in (0, 1):
                return bool(value)
        elif isinstance(value, str):
            if value.lower() in FALSE:
                return False
            elif value.lower() in TRUE:
                return True
        raise ValueError('Not a {}'.format(self.typename))

    def validate(self, value):
        self.convert(value)


class ListValue(ValueBase):
    type = list
    typename = 'list'

    def __init__(self, *args, options=None, **kwargs):
        self._options = options
        super().__init__(*args, **kwargs)

    @property
    def options(self):
        """tuple of allowed values or None to allow all values

        Calling `set` with a sequence that contains items not in `options`
        will raise a ValueError.
        """
        return self._options

    @options.setter
    def options(self, options):
        if isinstance(options, abc.Iterable):
            self._options = tuple(options)
        else:
            raise TypeError('options must be sequence, not {}: {!r}'
                            .format(type(options).__name__, options))

    def convert(self, value):
        if isinstance(value, str):
            return self.type(value.strip() for value in value.split(','))
        elif isinstance(value, abc.Iterable):
            return self.type(value)
        else:
            raise ValueError('Not a {}'.format(self.typename))

    def validate(self, lst):
        if self.options is not None:
            # Only items in self.options are allowed
            invalid_items = []
            for item in lst:
                if item not in self.options:
                    invalid_items.append(item)

            if invalid_items:
                raise ValueError('Invalid value{}: {}'.format(
                    's' if len(invalid_items) != 1 else '',
                    self.str(invalid_items)))

    def str(self, value=None, default=False):
        if default:
            lst = self.default
        elif value is not None:
            try:
                lst = self.convert(value)
            except ValueError:
                lst = str(value)
        else:
            lst = self.value
        return ', '.join(str(item) for item in lst)


class SetValue(ListValue):
    """ListValue with unique elements (order is preserved)"""
    type = list
    typename = 'set'

    def convert(self, value):
        lst = super().convert(value)
        # Make list items unique while preserving order
        seen = set()
        return [x for x in lst if not (x in seen or seen.add(x))]


class OptionValue(ValueBase):
    def __init__(self, *args, options=(), **kwargs):
        self._options = tuple(options)
        super().__init__(*args, **kwargs)

    @property
    def type(self):
        if self._options:
            return type(self._options[0])
        else:
            raise RuntimeError('Cannot guess value type: {!r}'.format(self))

    @property
    def typename(self):
        optvals = sorted(str(o) for o in self._options)
        return 'option: ' + ', '.join(optvals)

    def validate(self, value):
        if value not in self._options:
            optvals = sorted(str(o) for o in self._options)
            raise ValueError('Must be one of: {}'.format(', '.join(optvals)))

    def convert(self, value):
        self.validate(value)
        return value



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt


class ClientError(Exception):
    prefix = ''
    def __init__(self, message, *args, **kwargs):
        super().__init__(self.prefix + str(message))

    def __eq__(self, other):
        return type(self) == type(other) and self.args == other.args

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(type(self).__name__ + str(self))

class ConnectionError(ClientError):
    prefix = 'Connection failed: '

class RPCError(ClientError):
    prefix = 'Invalid RPC request: '

class AuthError(ClientError):
    prefix = 'Authentication failed: '

class URLParserError(ClientError):
    prefix = 'Invalid URL: '



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid
import collections


class TabID(int):
    def __repr__(self):
        return '<TabID %d>' % self


def _find_unused_id(existing_ids):
    """Find lowest unused ID in `existing_ids`"""
    if not existing_ids:
        return TabID(0)
    else:
        for id_candidate in range(0, max(existing_ids)+2):
            if id_candidate not in existing_ids:
                return TabID(id_candidate)
    raise RuntimeError('How did you get here?')


class Tabs(urwid.Widget):
    """Organize multiple widgets in tabs"""

    _sizing = frozenset([urwid.FLOW, urwid.BOX])

    def __init__(self, *contents, tabbar=None):
        """Create new Tabs widget

        contents: Iterable of dictionaries or iterables that match the
                  arguments of the `insert` method
        tabbar: Columns object that is used to display tab titles or any
                object with an 'original_widget' attribute (e.g. AttrMap) that
                returns a Columns object
        """
        if tabbar is None:
            self._tabbar = urwid.Columns([], dividechars=1)
            self._tabbar_render = self._tabbar.render
        elif hasattr(tabbar, 'original_widget'):
            self._tabbar = tabbar.original_widget
            self._tabbar_render = tabbar.render
        elif isinstance(tabbar, urwid.Columns):
            self._tabbar = tabbar
            self._tabbar_render = self._tabbar.render
        else:
            raise ValueError('tabbar must be Columns, not {}: {!r}'
                             .format(type(tabbar).__name__, tabbar))

        self._ids = []
        self._contents = urwid.MonitoredFocusList()
        for content in contents:
            if not isinstance(content, collections.Mapping):
                content = dict(zip(('title', 'widget', 'position', 'focus'),
                                   content))
            self.insert(**content)

    def render(self, size, focus=False):
        if len(size) < 2:
            cols, rows = (size[0], None)
        else:
            cols, rows = size

        if len(self._contents) < 1:
            # No contents - return empty canvas
            return urwid.SolidCanvas(' ', cols, rows)

        combinelist = []
        position = self._contents.focus

        # Render tab bar and add it to combinelist.  The tab bar is always
        # rendered as focused to highlight the focused tab.
        canvas = self._tabbar_render((cols,), True)
        combinelist.append((canvas, position, True))
        if rows is not None:
            rows -= 1  # Account for title bar

        # Render and add content of currently selected tab
        current_widget = self._contents[position]
        if current_widget is None:
            canvas = urwid.SolidCanvas(' ', cols, rows)
        else:
            if rows is not None:
                canvas = current_widget.render((cols,rows), focus)
            else:
                canvas = current_widget.render((cols,), focus)
        combinelist.append((canvas, position, focus))
        return urwid.CanvasCombine(combinelist)

    def get_index(self, position=None):
        """Return tab index at `position` or None if there are no tabs

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        if position is None:
            return self.focus_position
        elif isinstance(position, TabID):
            if position in self._ids:
                return self._ids.index(position)
            else:
                e = IndexError('No tab with ID: {}'.format(position))
                e.value = position
                raise e
        else:
            i = self.focus_position if position is None else position
            if i is not None and not 0 <= i < len(self._contents):
                e = IndexError('No tab at position: {}'.format(position))
                e.value = position
                raise e
        return i

    def get_id(self, position=None):
        """Return unique TabID of tab at `position` or None if there are no tabs

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        return self._ids[i] if i is not None else None

    def load(self, title, widget=None, position=None, focus=True):
        """Set content at `position`, in focused tab or in new tab

        If `position` is not None, is forwarded to `set_title`/`set_content`
        together with `title`/`widget`.

        If no tabs exist, a new tab is created with `title` and `widget`.  If
        `widget` is None, a blank widget is used.

        Otherwise, the focused tab's title and content is replaced with
        `set_title` and `set_content`.

        Set `focus` to False to load content in background.

        Return TabID of newly created tab
        """
        if position is not None:
            # Overload content in specified tab
            self.set_content(widget, position=position)
            self.set_title(title, position=position)
            if focus:
                if isinstance(position, TabID):
                    self.focus_id = position
                else:
                    self.focus_position = position
            return self.get_id(position)
        elif self.focus_position is None:
            # No tabs exist - create new tab
            return self.insert(title, widget, focus=focus)
        else:
            # Overload content in focused tab
            self.set_content(widget)
            self.set_title(title)
            return self.get_id()

    def insert(self, title, widget=None, position=-1, focus=True):
        """Insert new tab

        title: Any flow or fixed widget to use as the tab's title
        widget: Widget to show when this tab is selected or None
        position: Where to insert the new tab; int for list-like index or
                  'right'/'left' to insert next to focused tab
        focus: True to focus the new tab, False otherwise

        Return TabID of inserted tab
        """
        curpos = self.focus_position
        if position == 'right':
            newpos = (curpos+1) if curpos is not None else 0
        elif position == 'left':
            newpos = max(curpos, 0) if curpos is not None else 0
        elif isinstance(position, int):
            if position < 0:
                newpos = position + len(self._contents) + 1
            else:
                newpos = position
        else:
            raise ValueError('Invalid position: {!r}'.format(position))

        # Insert new tab ID
        this_id = _find_unused_id(self._ids)
        self._ids.insert(newpos, this_id)

        # Insert title
        options = self._tabbar.options('pack')
        self._tabbar.contents.insert(newpos, (title, options))

        # Insert content
        self._contents.insert(newpos, widget)
        if focus:
            self.focus_position = newpos
        return this_id

    def remove(self, position=None):
        """Remove tab `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        del self._ids[i]
        del self._tabbar.contents[i]
        del self._contents[i]

    def clear(self):
        """Remove all tabs"""
        while len(self._ids):
            self.remove(0)  # Remove tab at index 0

    def get_title(self, position=None):
        """Return tab title widget at `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        return self._tabbar.contents[i][0]

    def set_title(self, title, position=None):
        """Change the title widget of a tab

        title: New title widget
        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        self._tabbar.contents[i] = (title, self._tabbar.options('pack'))

    def get_content(self, position=None):
        """Return tab content widget at `position`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        return self._contents[i]

    def set_content(self, widget=None, position=None):
        """Set content of tab at `position` to `widget`

        position: Index (int), ID (TabID) or None (focused tab)

        Raises IndexError if tab can't be found.
        """
        i = self.get_index(position)
        self._contents[i] = widget

    @property
    def focus(self):
        """Content widget of currently focused tab or None if no tabs exist"""
        position = self._contents.focus
        if position is not None:
            return self._contents[position]
        return None

    @property
    def focus_position(self):
        """Index (starting from 0) of currently focused tab or None if no tabs exist"""
        return self._contents.focus

    @focus_position.setter
    def focus_position(self, position):
        if 0 <= position < len(self._contents):
            self._tabbar.contents.focus = position
            self._contents.focus = position
        else:
            raise IndexError('No tab at position: {!r}'.format(position))

    @property
    def focus_id(self):
        """TabID of currently focused tab or None if no tabs exist"""
        i = self.focus_position
        if i is not None:
            return self._ids[i]

    @focus_id.setter
    def focus_id(self, tabid):
        i = self.get_index(tabid)
        if 0 <= i < len(self._contents):
            self._tabbar.contents.focus = i
            self._contents.focus = i
        else:
            raise IndexError('No tab with ID: {}'.format(tabid))

    @property
    def contents(self):
        """Yields all content widgets"""
        for w in self._contents:
            yield w

    @property
    def titles(self):
        """Yields all tab title widgets"""
        for w in self._tabbar.contents:
            yield w[0]

    def __len__(self):
        return len(self._contents)

    def __iter__(self):
        return iter(self._contents)

    def selectable(self):
        return True

    def keypress(self, size, key):
        focus_widget = self.focus

        if focus_widget is not None:
            if len(size) > 1:
                #       maxcol   maxrow (-1 for the tab bar)
                size = (size[0], size[1]-1)
            else:
                size = (size[0],)

            if focus_widget.selectable():
                key = focus_widget.keypress(size, key)

        if key is not None:
            focus_pos = self.focus_position
            if key == 'left' and focus_pos > 0:
                self.focus_position -= 1
                key = None
            elif key == 'right' and focus_pos < len(self._contents)-1:
                self.focus_position += 1
                key = None
        return key



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Monkey patches that should be removed when they are resolved upstream"""

import urwid

# Workaround until this bug is resolved:
# https://github.com/urwid/urwid/pull/86
# https://github.com/urwid/urwid/issues/90
from distutils.version import LooseVersion
if LooseVersion(urwid.__version__) <= LooseVersion('1.3.1'):
    urwid.AsyncioEventLoop._idle_emulation_delay = 1/25


# Raise UnicodeDecodeError properly.
# https://github.com/urwid/urwid/pull/92
# https://github.com/urwid/urwid/pull/196
class AsyncioEventLoop_withfixedraise(urwid.AsyncioEventLoop):
    def _exception_handler(self, loop, context):
        exc = context.get('exception')
        if exc:
            loop.stop()
            if not isinstance(exc, urwid.ExitMainLoop):
                self._exc_info = exc
        else:
            loop.default_exception_handler(context)

    def run(self):
        """
        Start the event loop.  Exit the loop when any callback raises
        an exception.  If ExitMainLoop is raised, exit cleanly.
        """
        self._loop.set_exception_handler(self._exception_handler)
        self._loop.run_forever()
        if self._exc_info:
            raise self._exc_info

urwid.AsyncioEventLoop = AsyncioEventLoop_withfixedraise


# Tell ListBox to move to first/last item on 'home'/'end' keys.
class ListBox_with_home_end(urwid.ListBox):
    def keypress(self, size, key):
        key = super().keypress(size, key)
        if self.focus is not None:
            if key == 'home':
                self.focus_position = next(iter(self.body.positions()))
                self._invalidate()  # Don't know why this is needed?
                key = None
            elif key == 'end':
                self.focus_position = next(iter(self.body.positions(reverse=True)))
                self._invalidate()  # Don't know why this is needed?
                key = None
        return key

urwid.ListBox = ListBox_with_home_end



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Documentation commands"""

from ...logging import make_logger
log = make_logger(__name__)


from ..base import help as base
from .. import ExpectedResource
from . import make_tab_title


class HelpCmd(base.HelpCmdbase):
    provides = {'tui'}
    tui = ExpectedResource

    def display_help(self, topics, lines):
        import urwid
        titlew = make_tab_title(','.join(topics), 'tabs.help.unfocused', 'tabs.help.focused')
        lines = [urwid.Text(l) for l in lines]
        helpw = urwid.ListBox(urwid.SimpleListWalker(lines))
        self.tui.tabs.load(titlew, helpw)


class VersionCmd(base.VersionCmdbase):
    provides = {'tui'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Base classes for torrent commands"""

from ...logging import make_logger
log = make_logger(__name__)

import asyncio
from collections import abc

from .. import (InitCommand, ExpectedResource)
from . import mixin


class AddTorrentsCmdbase(metaclass=InitCommand):
    name = 'add'
    aliases = ('download','get')
    provides = set()
    category = 'torrent'
    description = 'Download torrents'
    usage = ('add <TORRENT> <TORRENT> <TORRENT> ... [<OPTIONS>]',)
    examples = ('add 72d7a3179da3de7a76b98f3782c31843e3f818ee',
                'add --stopped http://example.org/something.torrent')
    argspecs = (
        { 'names': ('TORRENT',), 'nargs': '+',
          'description': 'Link or path to torrent file, magnet link or hash' },
        { 'names': ('--stopped','-s'), 'action': 'store_true',
          'description': 'Do not start downloading the added torrent' },
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT, stopped):
        success = True
        force_torrentlist_update = False
        for source in TORRENT:
            response = await self.make_request(self.srvapi.torrent.add(source, stopped=stopped))
            success = success and response.success
            force_torrentlist_update = force_torrentlist_update or success

        # Update torrentlist AFTER all 'add' requests
        if force_torrentlist_update and hasattr(self, 'polling_frenzy'):
            self.polling_frenzy()
        return success


class AnnounceTorrentsCmdbase(metaclass=InitCommand):
    name = 'announce'
    aliases = ('an',)
    provides = set()
    category = 'torrent'
    description = 'Announce torrents to their trackers now if possible'
    usage = ('announce <TORRENT FILTER> <TORRENT FILTER> ...',)
    examples = ('announce tracker~example.org',)
    argspecs = (
        { 'names': ('TORRENT FILTER',), 'nargs': '*',
          'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            response = await self.make_request(
                self.srvapi.torrent.announce(tfilter),
                polling_frenzy=False)
            return response.success


def _make_SCRIPTING_doc(cmdname):
    return ( ("If invoked as a command line argument and the output does not "
              "go to a TTY (i.e. the terminal size can't be determined), "
              "the output is optimized for scripting.  Numbers are "
              "unformatted, columns are separated by a horizontal tab "
              "character ('\\t') and headers are not printed."),
             "",
             ("To enforce human-readable, formatted output, set the environment "
              "variables COLUMNS and LINES."),
             "",
             "\t$ \tCOLUMNS=80 LINES=24 {{APPNAME}} {CMDNAME} | less -R".format(CMDNAME=cmdname) )

def _make_SORT_ORDERS_doc(sortercls, option, setting, append=()):
    doc = [('The following sort orders can be specified with the {option} option '
            'or the "{setting}" setting:').format(option=option, setting=setting),
            '']

    for sname,s in sorted(sortercls.SORTSPECS.items()):
        snames = ', '.join((sname,) + s.aliases)
        doc.append('\t{}\t - \t{}'.format(snames, s.description))

    doc.extend(('',
                'Multiple sort orders are separated with "," without spaces.',
                '',
                'Sorting is reversed if the sort order is prepended by "!" or ".".',
                '',
                ('If "%s" is not given explicitly, it is always prepended to '
                 'the list of sort orders.') % sortercls.DEFAULT_SORT))
    if append:
        doc.extend(('',) + append)
    return tuple(doc)


def _make_COLUMNS_doc(columnspecs, option, setting, append=()):
    return (('The following columns can be specified with the {option} option '
             'or the "{setting}" setting:').format(option=option, setting=setting),
            '',
            '\t%s' % ', '.join(sorted(columnspecs)),
            '',
            'Columns are separated with "," without spaces.') \
            + append


class ListTorrentsCmdbase(mixin.get_torrent_sorter, mixin.get_tlist_columns,
                          metaclass=InitCommand):
    name = 'list'
    aliases = ('ls',)
    provides = set()
    category = 'torrent'
    description = 'List torrents'
    usage = ('list [<OPTIONS>]',
             'list [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('ls active',
                'ls !active',
                'ls seeds<10',
                'ls active&tracker~example.org',
                'ls active|idle&tracker~example')

    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`)'},

        { 'names': ('--sort', '-s'),
          'default_description': "current value of 'sort.torrents' setting",
          'description': ('Comma-separated list of sort orders '
                          "(see SORT ORDERS section)") },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'columns.torrents' setting",
          'description': ('Comma-separated list of column names '
                          "(see COLUMNS section)") },
    )

    from ...columns.tlist import COLUMNS
    from ...client.sorters.tsorter import TorrentSorter
    more_sections = {
        'COLUMNS': _make_COLUMNS_doc(COLUMNS, '--columns', 'columns.torrents'),
        'SORT ORDERS': _make_SORT_ORDERS_doc(TorrentSorter, '--sort', 'sort.torrents'),
        'SCRIPTING': _make_SCRIPTING_doc(name),
    }

    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, sort, columns):
        sort = self.cfg['sort.torrents'].value if sort is None else sort
        columns = self.cfg['columns.torrents'].value if columns is None else columns
        try:
            columns = self.get_tlist_columns(columns)
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=True,
                                           discover_torrent=False)
            sort = self.get_torrent_sorter(sort)
        except ValueError as e:
            log.error(e)
            return False
        else:
            log.debug('Listing %s torrents sorted by %s', tfilter, sort)
            if asyncio.iscoroutinefunction(self.make_tlist):
                return await self.make_tlist(tfilter, sort, columns)
            else:
                return self.make_tlist(tfilter, sort, columns)


class ListFilesCmdbase(mixin.get_flist_columns, metaclass=InitCommand):
    name = 'filelist'
    aliases = ('fls', 'lsf')
    provides = set()
    category = 'torrent'
    description = 'List files of torrent(s)'
    usage = ('filelist [<OPTIONS>]',
             'filelist [<OPTIONS>] [<TORRENT FILTER>] [<FILE FILTER>]')
    examples = ('filelist',
                "filelist A.Torrent.with.Files",
                "filelist A.Torrent.with.Files priority=low")
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},

        { 'names': ('FILE FILTER',), 'nargs': '?',
          'description': 'Filter expression (see `help filter`)' },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'columns.files' setting",
          'description': ('Comma-separated list of column names '
                          "(see COLUMNS section)") },
    )

    from ...columns.flist import COLUMNS
    more_sections = {
        'COLUMNS': _make_COLUMNS_doc(COLUMNS, '--columns', 'columns.files'),
        'SCRIPTING': _make_SCRIPTING_doc(name),
    }

    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, FILE_FILTER, columns):
        columns = self.cfg['columns.files'].value if columns is None else columns
        try:
            columns = self.get_flist_columns(columns)
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
            ffilter = self.select_files(FILE_FILTER,
                                        allow_no_filter=True,
                                        discover_file=False)
        except ValueError as e:
            log.error(e)
            return False

        log.debug('Listing %s files of %s torrents', ffilter, tfilter)

        if asyncio.iscoroutinefunction(self.make_flist):
            return await self.make_flist(tfilter, ffilter, columns)
        else:
            return self.make_flist(tfilter, ffilter, columns)


class ListPeersCmdbase(mixin.get_peer_sorter, mixin.get_plist_columns,
                       mixin.get_peer_filter, metaclass=InitCommand):
    name = 'peerlist'
    aliases = ('pls', 'lsp')
    provides = set()
    category = 'torrent'
    description = 'List connected peers of torrent(s)'
    usage = ('peerlist [<OPTIONS>]',
             'peerlist [<OPTIONS>] [<TORRENT FILTER>] [<PEER FILTER>]')
    examples = ('peerlist',
                "peerlist some_torrent",
                "peerlist some_torrent ip=127.0.0.1")
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},

        { 'names': ('PEER FILTER',), 'nargs': '?',
          'description': 'Filter expression (see `help filter`)' },

        { 'names': ('--sort', '-s'),
          'default_description': "current value of 'sort.peers' setting",
          'description': ('Comma-separated list of sort orders '
                          "(see SORT ORDERS section)") },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'columns.peers' setting",
          'description': ('Comma-separated list of column names '
                          "(see COLUMNS section)") },
    )

    from ...columns.plist import COLUMNS
    from ...client.sorters.psorter import TorrentPeerSorter
    more_sections = {
        'COLUMNS': _make_COLUMNS_doc(COLUMNS, '--columns', 'columns.peers', append=(
            '',
            'The "name" column is added automatically if multiple '
            'torrents could be listed potentially.')),
        'SORT ORDERS': _make_SORT_ORDERS_doc(TorrentPeerSorter, '--sort', 'sort.peers'),
        'SCRIPTING': _make_SCRIPTING_doc(name),
    }

    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, PEER_FILTER, sort, columns):
        columns = self.cfg['columns.peers'].value if columns is None else columns
        sort = self.cfg['sort.peers'].value if sort is None else sort
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=True,
                                           discover_torrent=True)
            pfilter = self.get_peer_filter(PEER_FILTER)
            sort    = self.get_peer_sorter(sort)
            columns = self.get_plist_columns(columns)
        except ValueError as e:
            log.error(e)
            return False

        # Unless we're listing peers of exactly one torrent, specified by its
        # ID, automatically add the 'name' column.
        if 'name' not in columns and \
           (not isinstance(tfilter, abc.Sequence) or len(tfilter) != 1):
            columns.insert(0, 'name')

        log.debug('Listing %s peers of %s torrents', pfilter, tfilter)

        if asyncio.iscoroutinefunction(self.make_plist):
            return await self.make_plist(tfilter, pfilter, sort, columns)
        else:
            return self.make_plist(tfilter, pfilter, sort, columns)


class MoveTorrentsCmdbase(metaclass=InitCommand):
    name = 'move'
    aliases = ('mv',)
    provides = set()
    category = 'torrent'
    description = "Change torrents' location"
    usage = ('move [<TORRENT FILTER>] <PATH>',)
    examples = ('move new/path',
                'move size>50g path/to/lots/of/storage')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},

        {'names': ('PATH',),
         'description': ('New location of the specified torrent(s).  If PATH is relative '
                         '(does not start with "/"), it is relative to the value of the '
                         'setting "srv.path.complete".')},
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, PATH):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            response = await self.make_request(self.srvapi.torrent.move(tfilter, PATH),
                                               polling_frenzy=True)
            return response.success



class PriorityCmdbase(metaclass=InitCommand):
    name = 'priority'
    aliases = ('prio',)
    provides = set()
    category = 'torrent'
    description = 'Change download priority of files'
    usage = ('priority <PRIORITY>',
             'priority <PRIORITY> [<TORRENT FILTER>] [<FILE FILTER>]')
    examples = ('priority low',
                'priority high "some torrent" size>12M')
    argspecs = (
        { 'names': ('PRIORITY',),
          'description': ("File priority; must be low/normal/high/shun, "
                          "l/n/h/s or -/./+/0")},
        {'names': ('TORRENT FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
        { 'names': ('FILE FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused file in the TUI'},
    )

    srvapi = ExpectedResource

    _PRIORITY = {'l': 'low',    '-': 'low',    'low': 'low',
                 'n': 'normal', '.': 'normal', 'normal': 'normal',
                 'h': 'high',   '+': 'high'  , 'high': 'high',
                 's': 'shun',   '0': 'shun',   'shun': 'shun'}

    async def run(self, PRIORITY, TORRENT_FILTER, FILE_FILTER):
        try:
            priority = self._PRIORITY[PRIORITY.lower()]
        except KeyError:
            log.error('Invalid priority: {!r}'.format(PRIORITY))
            return False

        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
            ffilter = self.select_files(FILE_FILTER,
                                        allow_no_filter=True,
                                        discover_file=True)
        except ValueError as e:
            log.error(e)
            return False

        if not isinstance(tfilter, tuple):
            # tfilter must be TorrentFilter instance, which means the user
            # specified a filter and will be informed about the matches.
            if isinstance(ffilter, tuple):
                # The user did specify a torrent filter but not a file filter,
                # so select_files() may have returned a focused file.  But we
                # assume that the user meant all files of the matching
                # torrents, otherwise they wouldn't have given a torrent
                # filter.
                ffilter = None

            msg = 'New download priority of %s files in %s torrents: %s' % (
                'all' if ffilter is None else ffilter, tfilter, priority)
            log.info(msg)
            quiet = False
        else:
            # We're operating on the focused file and success is indiciated by
            # the updated file list, so no info message necessary.
            quiet = True

        # If ffilter is a tuple, it is a tuple of (torrent_id, file_id) tuples.
        # In that case, we overload tfilter with the torrent_ids from ffilter.
        if isinstance(ffilter, tuple):
            tfilter = tuple(set(ids[0] for ids in ffilter))
        log.debug('Setting file download priority to %s for %s files of %s torrents',
                  priority, ffilter, tfilter)

        response = await self.make_request(
            self.srvapi.torrent.file_priority(priority, tfilter, ffilter),
            polling_frenzy=True, quiet=quiet)
        return response.success


class RemoveTorrentsCmdbase(metaclass=InitCommand):
    name = 'remove'
    aliases = ('rm', 'delete')
    provides = set()
    category = 'torrent'
    description = 'Remove torrents'
    usage = ('remove [<OPTIONS>]',
             'remove [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('remove',
                'remove "some torrent" another\ torrent and_this_torrent',
                'remove -d "unwanted torrent"')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},

        { 'names': ('--delete-files','-d'), 'action': 'store_true',
          'description': 'Delete any downloaded files' },
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, delete_files):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            response = await self.make_request(
                self.srvapi.torrent.remove(tfilter, delete=delete_files),
                polling_frenzy=True)
            return response.success


# Argument definitions that are shared between commands
ARGSPEC_TOGGLE = {
    'names': ('--toggle','-t'), 'action': 'store_true',
    'description': ('Start TORRENT if stopped and vice versa')
}

class StartTorrentsCmdbase(metaclass=InitCommand):
    name = 'start'
    aliases = ()
    provides = set()
    category = 'torrent'
    description = 'Start downloading torrents'
    usage = ('start [<OPTIONS>]',
             'start [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('start',
                "start 'night of the living dead' Metropolis",
                'start ubuntu --force')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
        { 'names': ('--force','-f'), 'action': 'store_true',
          'description': 'Ignore download queue' },
        ARGSPEC_TOGGLE,
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, toggle, force):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            if toggle:
                response = await self.make_request(
                    self.srvapi.torrent.toggle_stopped(tfilter, force=force),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    self.srvapi.torrent.start(tfilter, force=force),
                    polling_frenzy=True)
            return response.success


class StopTorrentsCmdbase(metaclass=InitCommand):
    name = 'stop'
    aliases = ('pause',)
    provides = set()
    category = 'torrent'
    description = 'Stop downloading torrents'
    usage = ('stop [<OPTIONS>]',
             'stop [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('stop',
                'stop "night of the living dead" idle',
                'stop --toggle ubuntu')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
        ARGSPEC_TOGGLE,
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, toggle):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            if toggle:
                response = await self.make_request(
                    self.srvapi.torrent.toggle_stopped(tfilter),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    self.srvapi.torrent.stop(tfilter),
                    polling_frenzy=True)
            return response.success


class VerifyTorrentsCmdbase(metaclass=InitCommand):
    name = 'verify'
    aliases = ('check',)
    provides = set()
    category = 'torrent'
    description = 'Verify downloaded torrent data'
    usage = ('verify [<OPTIONS>]',
             'verify [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('verify',
                'verify debian')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            response = await self.make_request(self.srvapi.torrent.verify(tfilter),
                                               polling_frenzy=False)
            return response.success



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filter sequences of Torrents, TorrentFiles by arbitrary criteria"""

from ...logging import make_logger
log = make_logger(__name__)

import operator
import re
from collections import abc
from itertools import zip_longest


class BoolFilterSpec():
    """Boolean filter specification"""

    def __init__(self, func, needed_keys=(), aliases=(), description='No description'):
        self.filter_function = func
        self.needed_keys = needed_keys
        self.aliases = aliases
        self.description = description


class CmpFilterSpec(BoolFilterSpec):
    """Comparative filter specification"""

    def __init__(self, *args, value_type, **kwargs):
        super().__init__(*args, **kwargs)
        self.value_type = value_type

    def make_filter_func(self, operator, value):
        func = self.filter_function
        return lambda obj: func(obj, operator, value)


class Filter():
    """Match sequences of objects against a single filter"""

    _OPERATORS = {
        '=': operator.__eq__, '~': operator.__contains__,
        '>': operator.__gt__, '<': operator.__lt__,
        '>=': operator.__ge__, '<=': operator.__le__,
    }
    _OP_CHARS = ''.join(_OPERATORS)
    _OP_LIST = '(?:' + '|'.join(sorted(_OPERATORS, key=lambda op: len(op), reverse=True)) + ')'
    _INVERT_CHAR = '!'
    _FILTER_REGEX = re.compile(r'^'
                               r'(?P<invert1>' + _INVERT_CHAR + '?)'
                               r'(?P<name>[^' + _OP_CHARS+_INVERT_CHAR + ']*)'
                               r'(?P<invert2>' + _INVERT_CHAR + '?)'
                               r'(?P<op>' + _OP_LIST + '|)'
                               r'(?P<value>.*)$')

    DEFAULT_FILTER = None

    BOOLEAN_FILTERS = {}
    COMPARATIVE_FILTERS = {}

    @classmethod
    def _check_value(cls, name, value, op):
        """Convert `value` to correct type for comparative filter `name`

        Also ensure operator `op` is compatible with `value`.

        Raises ValueError
        """
        if name not in cls.COMPARATIVE_FILTERS:
            return value

        # Convert value to proper type
        target_type = cls.COMPARATIVE_FILTERS[name].value_type
        if type(value) is not target_type:
            try:
                value = target_type(value)
            except ValueError:
                raise ValueError('Invalid value for filter {!r}: {!r}'.format(name, value))

        # Make sure value and operator are compatible
        if op == '~' and not isinstance(value, str):
            raise ValueError('Invalid operator for filter {!r}: {}'.format(name, op))

        return value

    @classmethod
    def _resolve_alias(cls, name):
        if not hasattr(cls, '_aliases'):
            aliases = {}
            for fseq in (cls.BOOLEAN_FILTERS, cls.COMPARATIVE_FILTERS):
                for fname,f in fseq.items():
                    for a in f.aliases:
                        aliases[a] = fname
            cls._aliases = aliases
        return cls._aliases.get(name, name)

    def __init__(self, filter_str=''):
        # name: Name of filter (user-readable string)
        # invert: Whether to invert filter (bool)
        # op: Comparison operator as string (see _OPERATORS)
        # value: User-given value as string (will be converted to proper type)
        name, invert, op, value = (None, False, None, None)
        if filter_str != '':
            match = self._FILTER_REGEX.fullmatch(filter_str)
            if match is None:
                raise ValueError('Invalid filter: {!r}'.format(filter_str))
            else:
                name = match.group('name')
                op = match.group('op') or None
                invert = bool(match.group('invert1')) ^ bool(match.group('invert2'))
                value = match.group('value')
                value = None if value.strip() == '' else value

        # No operator but a value doesn't make any sense
        if op is None and value is not None:
            raise ValueError('Malformed filter expression: {!r}'.format(filter_str))

        # Handle spaces around operator: If there's a space before the
        # operator, strip value.  Otherwise, preserve them.  In any case,
        # strip name.
        if op is not None:
            if name.endswith(' '):
                value = value.strip(' ')
        if name is not None:
            name = name.strip()

        if name is None:
            # No filter_str provided
            name = 'all'
        elif name is '':
            # No filter name is given, but a value and maybe an operator.
            name = self.DEFAULT_FILTER
        else:
            name = self._resolve_alias(name)

        # Make sure value has the correct type and operator is compatible
        if value is not None:
            value = self._check_value(name, value, op)

        log.debug('Parsed filter %r: name=%r, invert=%r, op=%r, value=%r',
                  filter_str, name, invert, op, value)

        # Filter that doesn't use value argument
        if name in self.BOOLEAN_FILTERS:
            f = self.BOOLEAN_FILTERS[name]
            self._filter_func = f.filter_function
            self._needed_keys = f.needed_keys

        # Filter that needs an argument
        elif name in self.COMPARATIVE_FILTERS:
            f = self.COMPARATIVE_FILTERS[name]
            self._needed_keys = f.needed_keys
            if op is None and value is None:
                # Abuse comparative filter as boolean filter
                # (e.g. 'peers-connected' matches torrents with peers-connected!=0)
                self._filter_func = lambda obj, keys=f.needed_keys: all(bool(obj[key]) for key in keys)
            elif op is None:
                ops = '[' + '|'.join(sorted(self._OPERATORS)) + ']'
                raise ValueError('Missing operator and value: {} {} ...'.format(name, ops))
            elif value is None:
                raise ValueError('Missing value: {} ...'.format(filter_str))
            else:
                self._filter_func = f.make_filter_func(self._OPERATORS[op], value)

        elif value is op is None and self.DEFAULT_FILTER is not None:
            # `name` is no known filter - default to DEFAULT_FILTER with operator '~'.
            value = name
            op = '~'
            name = self.DEFAULT_FILTER
            if name in self.BOOLEAN_FILTERS:
                f = self.BOOLEAN_FILTERS[name]
                self._filter_func = f.filter_function
            elif name in self.COMPARATIVE_FILTERS:
                f = self.COMPARATIVE_FILTERS[name]
                self._filter_func = f.make_filter_func(self._OPERATORS[op], value)
            else:
                raise RuntimeError('Default filter {!r} does not exist: {!r}'
                                   .format(name, ', '.join(tuple(self.BOOLEAN_FILTERS) +
                                                           tuple(self.COMPARATIVE_FILTERS))))
            self._needed_keys = f.needed_keys

        else:
            raise ValueError('Invalid filter name: {!r}'.format(name))

        self._name, self._invert, self._op, self._value = name, invert, op, value
        self._hash = hash((name, invert, op, value))

    def apply(self, objs, invert=False, key=None):
        """Yield matching objects or `key` of each matching object"""
        invert = self._invert ^ bool(invert)  # xor
        wanted = self._filter_func
        for i in objs:
            if wanted(i) ^ invert:
                yield i if key is None else i[key]

    def match(self, obj):
        """Return True if `obj` matches, False otherwise"""
        return self._filter_func(obj) ^ self._invert

    def __str__(self):
        if self._name is None:
            return 'all'
        elif self._value is None:
            return ('!' if self._invert else '') + self._name
        else:
            name = self._name if self._name != self.DEFAULT_FILTER else ''
            op = ('!' if self._invert else '') + self._op
            val = str(self._value)
            return name + op + val

    @property
    def needed_keys(self):
        return self._needed_keys

    def __eq__(self, other):
        if isinstance(other, type(self)):
            for attr in ('_name', '_value', '_invert', '_op'):
                if getattr(self, attr) != getattr(other, attr):
                    return False
            return True
        else:
            return NotImplemented

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, str(self))

    def __hash__(self):
        return self._hash



class FilterChain():
    """One or more filters combined with AND and OR operators"""

    filterclass = None
    _op_regex = re.compile(r'([&|])')

    def __init__(self, filters=''):
        if not isinstance(self.filterclass, type) or not issubclass(self.filterclass, Filter):
            raise RuntimeError('Attribute "filterclass" must be set to a Filter class, not {!r}'
                               .format(self.filterclass))

        if isinstance(filters, str):  # Because str is also instance of abc.Sequence
            pass
        elif isinstance(filters, abc.Sequence) and all(isinstance(f, str) for f in filters):
            filters = '|'.join(filters)
        elif not isinstance(filters, str):
            raise TypeError('filters must be string or sequence of strings, not {}: {!r}'
                            .format(type(filters).__name__, filters))

        # self._filterchains is a tuple of tuples.  Each inner tuple combines
        # filters with AND.  The outer tuple combines the inner, AND-combined
        # tuples with OR.
        parts = tuple(part for part in self._op_regex.split(filters) if part is not '')
        if len(parts) < 1:
            self._filterchains = ()
        else:
            if parts[0] in '&|':
                raise ValueError('Filter can\'t start with operator: {!r}'.format(parts[0]))
            elif parts[-1] in '&|':
                raise ValueError('Filter can\'t end with operator: {!r}'.format(parts[-1]))

            filters = []
            ops = []
            expect = 'filter'
            nofilter = self.filterclass()
            for i,part in enumerate(parts):
                if expect is 'filter':
                    if part not in '&|':
                        f = self.filterclass(part)
                        if f == nofilter:
                            # part is something like 'all' or '*' - this
                            # disables all other filters
                            filters = []
                            ops = []
                            break
                        else:
                            filters.append(f)
                            expect = 'operator'
                            continue
                elif expect is 'operator':
                    if part in '&|':
                        ops.append(part)
                        expect = 'filter'
                        continue
                raise ValueError('Consecutive operators: {!r}'.format(''.join(parts[i-2:i+2])))

            if filters:
                fchain = [[]]
                for filter,op in zip_longest(filters, ops):
                    fchain[-1].append(filter)
                    if op is '|':
                        fchain.append([])
                self._filterchains = tuple(tuple(x) for x in fchain)
            else:
                self._filterchains = ()

    def apply(self, objects):
        """Yield matching objects from iterable `objects`"""
        if self._filterchains:
            yield from filter(self.match, objects)
        else:
            yield from objects

    def match(self, obj):
        """Whether `obj` matches this filter chain"""
        # All filters in an AND_chain must match for the AND_chain to
        # match.  At least one AND_chain must match.
        if len(self._filterchains) < 1:
            return True
        else:
            return any(all(f.match(obj) for f in AND_chain)
                       for AND_chain in self._filterchains)

    @property
    def needed_keys(self):
        """The object keys needed for filtering"""
        keys = set()
        for chain in self._filterchains:
            for filter in chain:
                keys.update(filter.needed_keys)
        return tuple(keys)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            # Compare sets because order doesn't matter (foo&bar|baz is the
            # same as baz|bar&foo). Use frozensets because sets are not
            # hashable.
            self_fc_sets = set(frozenset(x) for x in self._filterchains)
            other_fc_sets = set(frozenset(x) for x in other._filterchains)
            return self_fc_sets == other_fc_sets

    def __str__(self):
        if len(self._filterchains) < 1:
            return 'all'
        else:
            OR_chains = []
            for AND_chain in self._filterchains:
                OR_chains.append('&'.join(str(f) for f in AND_chain))
            return '|'.join(OR_chains)

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, str(self))

    def __add__(self, other):
        cls = type(self)
        nofilter = cls()
        if not isinstance(other, cls):
            return NotImplemented
        elif other == nofilter or self == nofilter:
            return nofilter
        else:
            # Start with our own stuff
            new_fc = list(self._filterchains)

            # Because foo&bar is the same as bar&foo, comparing sets makes
            # everything much easier
            self_fc_sets = tuple(set(x) for x in self._filterchains)

            # Copy each AND_chain from other unless we already have it
            for AND_chain in other._filterchains:
                AND_chain_set = set(AND_chain)
                if AND_chain_set not in self_fc_sets:
                    new_fc.append(AND_chain)

            # Make string from new_fc
            OR_chains = []
            for AND_chain in new_fc:
                OR_chains.append('&'.join(str(f) for f in AND_chain))
            new_fc_str = '|'.join(OR_chains)

            return cls(new_fc_str)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import logging
import sys

MAIN_PACKAGE_NAME = __name__[:__name__.index('.')]

def make_logger(modname=None):
    if modname is None:
        # Default logger, for which debugging is always enabled if it is for
        # any other modules
        return logging.getLogger(MAIN_PACKAGE_NAME)
    elif modname.startswith(MAIN_PACKAGE_NAME+'.'):
        # Remove main package name from logger name ('foo.bar.baz' -> 'bar.baz')
        loggername = modname[len(MAIN_PACKAGE_NAME)+1:]
    else:
        loggername = modname
    return logging.getLogger(loggername)

log = make_logger()

def setup(debugmods, filepath=None):
    class PerLevelFormatter(logging.Formatter):
        """Use different formatter per level"""
        datefmt = '%H:%M:%S'
        formatters = {
            logging.DEBUG: logging.Formatter('%(asctime)s: [%(name)s] %(message)s', datefmt=datefmt),
        }
        default_fmtr = logging.Formatter('%(message)s', datefmt=datefmt)

        def format(self, record):
            fmtr = self.formatters.get(record.levelno, self.default_fmtr)
            return fmtr.format(record)

    root_logger = logging.getLogger()
    formatter = PerLevelFormatter()
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    root_logger.addHandler(stream_handler)
    root_logger.setLevel(logging.INFO)

    if debugmods:
        # Enable debugging for default logger
        logging.getLogger(MAIN_PACKAGE_NAME).setLevel(logging.DEBUG)

        modnames = debugmods
        for modname in modnames:
            logging.getLogger(modname).setLevel(logging.DEBUG)
            log.debug('Debugging messages enabled: ' + modname)

    if filepath is not None:
        file_handler = logging.FileHandler(filepath)
        file_handler.setFormatter(formatter)
        root_logger.addHandler(file_handler)
        log.debug('Logging to file: %r', file_handler.baseFilename)


def redirect_level(level, stream=sys.stderr):
    root_logger = logging.getLogger()

    # Add filters to all existing handlers to NOT log the specified level
    for h in root_logger.handlers:
        h.addFilter(lambda record: record.levelname != level)

    # Add new handler that only logs the specified level
    lvlhandler = logging.StreamHandler(stream)
    lvlhandler.addFilter(lambda record: record.levelname == level)
    root_logger.addHandler(lvlhandler)


def enable_profiling(filepath):
    import cProfile, signal
    p = cProfile.Profile()
    p.enable()
    def stop_profiling(signum, frame):
        p.disable()
        p.dump_stats(filepath)
        sys.exit(1)
    signal.signal(signal.SIGTERM, stop_profiling)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import urwid
from collections import abc

from .. import main as tui

from ..table import Table
from .tlist_columns import TUICOLUMNS

COLUMNS_FOCUS_MAP = {}
for col in TUICOLUMNS.values():
    COLUMNS_FOCUS_MAP.update(col.style.focus_map)


class TorrentListItemWidget(urwid.WidgetWrap):
    def __init__(self, torrent, cells):
        self._torrent = torrent
        self._tid = torrent['id']
        self._cells = cells
        self.update(torrent)
        row = urwid.AttrMap(urwid.AttrMap(cells, attr_map=None, focus_map=COLUMNS_FOCUS_MAP),
                            'torrentlist', 'torrentlist.focused')
        super().__init__(row)

    def update(self, torrent):
        for widget in self._cells.widgets:
            widget.update(torrent)
        self._torrent = torrent

    @property
    def torrent_id(self):
        return self._tid

    @property
    def torrent(self):
        return self._torrent

    @property
    def is_marked(self):
        return self._cells.marked.is_marked

    @is_marked.setter
    def is_marked(self, is_marked):
        self._cells.marked.is_marked = bool(is_marked)


class TorrentListWidget(urwid.WidgetWrap):
    def __init__(self, sort=None, tfilter=None, columns=[]):
        self._sort = sort
        self._tfilter = tfilter
        self._columns = columns
        self.title_updater = None

        self._table = Table(**TUICOLUMNS)
        self._table.columns = columns

        self._torrents = ()
        self._walker = urwid.SimpleFocusListWalker([])
        self._listbox = urwid.ListBox(self._walker)
        self._marked = set()

        pile = urwid.Pile([
            ('pack', urwid.AttrMap(self._table.headers, 'torrentlist.header')),
            self._listbox
        ])
        super().__init__(urwid.AttrMap(pile, 'torrentlist'))

        self._register_request()

    def _register_request(self):
        # Get keys needed for sort order, tfilter and columns
        keys = []
        if self._sort is not None:
            keys.extend(self._sort.needed_keys)
        if hasattr(self._tfilter, 'needed_keys'):
            keys.extend(self._tfilter.needed_keys)
        for colname in self._columns:
            keys.extend(TUICOLUMNS[colname].needed_keys)

        # Register new request in request pool
        log.debug('Registering keys for %s: %s', self.id, keys)
        tui.srvapi.treqpool.register(self.id, self._handle_tlist,
                                     keys=keys,
                                     tfilter=self._tfilter)
        tui.srvapi.treqpool.poll()

    def _handle_tlist(self, torrents):
        self._torrents = tuple(torrents)
        if self.title_updater is not None:
            self.title_updater(self.title)
        self._invalidate()

    def render(self, size, focus=False):
        if self._torrents is not None:
            self._update_listitems()
            self._torrents = None
        return super().render(size, focus)

    def _update_listitems(self):
        # Remember focused torrent
        focused_torrent = self.focused_torrent

        walker = self._walker
        tdict = {t['id']:t for t in self._torrents}
        dead_tws = []
        for tw in walker:  # tw = TorrentListItemWidget
            tid = tw.torrent_id
            try:
                # Update existing torrent widget with new data
                tw.update(tdict[tid])
                del tdict[tid]
            except KeyError:
                # Torrent has been removed
                dead_tws.append(tw)

        # Remove list items that don't exist in self._torrents anymore
        marked = self._marked
        for tw in dead_tws:
            walker.remove(tw)
            marked.discard(tw)  # List items are also stored in self._marked

        # Any torrents that haven't been used to update an existing torrent are new
        if tdict:
            widgetcls = tui.keymap.wrap(TorrentListItemWidget, context='torrent')
            for tid in tdict:
                self._table.register(tid)
                row = self._table.get_row(tid)
                walker.append(widgetcls(tdict[tid], row))

        # Sort torrents
        if self._sort is not None:
            self._sort.apply(walker,
                            item_getter=lambda tw: tw.torrent,
                            inplace=True)

        # If necessary, re-focus previously focused torrent
        if focused_torrent is not None and self.focused_torrent is not None and \
           focused_torrent.torrent_id != self.focused_torrent.torrent_id:
            focused_tid = focused_torrent.torrent_id
            for i,tw in enumerate(walker):
                if tw.torrent_id == focused_tid:
                    self._listbox.focus_position = i
                    break

    def clear(self):
        """Remove all list items"""
        for tw in tuple(self._walker):
            tw.torrent.clearcache()
            self._walker.remove(tw)
        tui.srvapi.treqpool.poll()

    def update(self):
        """Call `clear` and then poll immediately"""
        self.clear()
        tui.srvapi.treqpool.poll()

    def set(self, columns=None, tfilter=None, sort=None):
        update_request = False
        old_id = self.id
        if columns is not None:
            self._table.columns = columns
            update_request = True
        if tfilter is not None:
            self._tfilter = tfilter
            update_request = True
        if sort is not None:
            self._sort = sort
            update_request = True
        if update_request:
            tui.srvapi.treqpool.remove(old_id)
            self._register_request()

    @property
    def title(self):
        if self._tfilter is None:
            title = ['<all>']
        elif isinstance(self._tfilter, abc.Sequence):
            title = ['<handpicked>']
        else:
            title = [str(self._tfilter)]
        sortstr = str(self._sort)
        if sortstr is not self._sort.DEFAULT_SORT:
            title.append('{%s}' % sortstr)

        # If this method was called before rendering, the contents of the
        # listbox widget are inaccurate and we have to use self._torrents.
        # But if we're called after rendering, self._torrents is reset to
        # None.
        if self._torrents is not None:
            title.append('[%d]' % len(self._torrents))
        else:
            title.append('[%d]' % len(self._listbox.body))

        return ' '.join(title)

    @property
    def id(self):
        return 'tlist%s:%s' % (id(self), self.title)

    @property
    def focused_torrent(self):
        return self._listbox.focus

    @property
    def focus_position(self):
        return self._listbox.focus_position

    @focus_position.setter
    def focus_position(self, focus_position):
        self._listbox.focus_position = min(focus_position, len(self._listbox.body)-1)

    def mark(self, toggle=False, all=False):
        """Mark the currently focused item or all items"""
        self._set_mark(True, toggle=toggle, all=all)

    def unmark(self, toggle=False, all=False):
        """Unmark the currently focused item or all items"""
        self._set_mark(False, toggle=toggle, all=all)

    @property
    def marked(self):
        """Generator that yields TorrentListItemWidgets"""
        yield from self._marked

    def _set_mark(self, mark, toggle=False, all=False):
        if toggle and self.focused_torrent is not None:
            mark = not self.focused_torrent.is_marked

        for widget in self._select_items_for_marking(all):
            widget.is_marked = mark
            if mark:
                self._marked.add(widget)
            else:
                self._marked.discard(widget)

    def _select_items_for_marking(self, all):
        if self.focused_torrent is not None:
            if all:
                yield from self._listbox.body
            else:
                yield self.focused_torrent

    def refresh_marks(self):
        """Redraw the "marked" column in all rows"""
        for widget in self._listbox.body:
            widget.is_marked = widget.is_marked



# Create long_description from README.org if possible
long_description = ''
try:
    import pypandoc
    import os
    if os.path.exists('README.org'):
        long_description = pypandoc.convert('README.org', 'rst')
except ImportError:
    pass


with open('stig/version.py') as f:
    exec(f.read())  # Load __version__ into globals()

from setuptools import setup, find_packages
setup(
    name             = 'stig',
    version          = __version__,
    license          = 'GPLv3+',
    author           = 'Random User',
    author_email     = 'rndusr@posteo.de',

    description      = 'TUI and CLI client for the Transmission daemon',
    long_description = long_description,

    url              = 'https://github.com/rndusr/stig',
    keywords         = 'bittorrent torrent transmission',

    packages         = find_packages(),
    package_data     = {'stig': ['settings/default.theme']},

    install_requires = [
        'urwid>=1.3.0',
        'aiohttp>=0.22.5',
        'urwidtrees>=1.0.3dev0',
        'appdirs',
        'blinker',
    ],

    entry_points = { 'console_scripts': [ 'stig = stig:run' ] },

    classifiers = [
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Environment :: Console',
        'Operating System :: Unix',
        'Development Status :: 3 - Alpha',
    ],
)
